# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::CGI::Index;

use strict;
use warnings;

use base qw( AdvPowerMgr::CGI );

use JSON qw( to_json );

use AdvPowerMgr::DropPrivileges;
use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;

our $msgtable = {};

sub header {
    my $class = shift;

    return <<EOF;
Content-Type: text/html

EOF
}

sub code {
    my $class = shift;

    #
    # check_privilege() requires to be root, so drop privileges just after this call
    #
    my ($token, $user, $is_admin) = AdvPowerMgr::System->check_privilege();

    AdvPowerMgr::DropPrivileges->drop();

    info("SynoToken='%s', User '%s' (%s)", $token, $user, $is_admin ? "administrator" : "regular user");

    my $params = {};
    $params->{pattern} = qr/^ui\./ if $is_admin;

    my $table = AdvPowerMgr::I18N->get($params) or die_ie("LANG");
    $msgtable = $table->{msgtable};

    if (!$is_admin) {
        info("User is not administrator, refusing to load UI");
        my $html = <<EOF;
<!DOCTYPE HTML>
<html><head><meta charset="UTF-8"><title>$msgtable->{'ui.APM'}</title></head>
<body><h1>$msgtable->{'auth_error.title'}</h1><p>$msgtable->{'auth_error.text'}</p></body>
</html>
EOF
        $class->html($html);
        return;
    }

    my $json = to_json($msgtable);

    my $locale;
    opendir(my $dh, 'resources/locale') or die_open('resources/locale');
    my $regex = qr/^ext-lang-([a-zA-Z0-9]+(?:_[a-zA-Z0-9]+)?)\.js$/;
    my @available = ();
    while (my $f = readdir($dh)) {
        next unless $f =~ $regex;
        my $f = $1;
        $f =~ s#_#-#;
        push(@available, $f);
    }
    closedir($dh);
    $locale = AdvPowerMgr::I18N->bestlang(@available);
    die_ie("No language file found") unless defined($locale);
    $locale =~ s#-#_#;
    $locale = "resources/locale/ext-lang-$locale.js";
    info("Using ExtJS locale file '%s'", $locale);

    my $version = AdvPowerMgr::System->appversion();
    info("AdvPowerMgr version is '%s'", $version);

    info("Building HTML page to load UI");
    my $html = <<EOF;
<!DOCTYPE HTML>
<html>
<head>
    <meta charset="UTF-8">
    <title>$msgtable->{'ui.APM'}</title>
    <link rel="stylesheet" href="resources/APM-all.css?version=$version"/>
    <style type="text/css">
        html, body {
            height: 100%;
            background-color: #1985D0
        }

        #appLoadingIndicator {
            position: absolute;
            top: 50%;
            margin-top: -15px;
            text-align: center;
            width: 100%;
            height: 30px;
            -webkit-animation-name: appLoadingIndicator;
            -webkit-animation-duration: 0.5s;
            -webkit-animation-iteration-count: infinite;
            -webkit-animation-direction: linear;
        }

        #appLoadingIndicator > * {
            background-color: #FFFFFF;
            display: inline-block;
            height: 30px;
            -webkit-border-radius: 15px;
            margin: 0 5px;
            width: 30px;
            opacity: 0.8;
        }

        @-webkit-keyframes appLoadingIndicator{
            0% {
                opacity: 0.8
            }
            50% {
                opacity: 0
            }
            100% {
                opacity: 0.8
            }
        }
    </style>
</head>
<body>
    <div id="appLoadingIndicator">
        <div></div>
        <div></div>
        <div></div>
    </div>
</body>
<script type="text/javascript">
    window.APMversion = "$version";
    window.SYNOtoken = "$token";
</script>
<script type="text/javascript" src="app.js?version=$version"></script>
<script type="text/javascript" src="$locale"></script>
</html>
EOF
    $class->html($html);
}

sub die_code {
    my $class = shift;
    my $error = shift or die_param('error');

    my $doctitle = $msgtable->{'ui.APM'} || "Advanced Power Manager";
    my $title = $msgtable->{'internal_error.title'} || "Internal error!";
    my $text = $msgtable->{'internal_error.text'} || "Please contact the development team!<br/>msg = $error";
    my $html = <<EOF;
<!DOCTYPE HTML>
<html><head><meta charset="UTF-8"><title>$doctitle</title></head>
<body><h1>$title</h1><p>$text</p></body>
</html>
EOF
    $class->html($html);
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
