# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Config;

use strict;
use warnings;

use base qw( AdvPowerMgr::Config::Component );

use File::Path qw( mkpath );

use JSON qw( to_json from_json );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;

our %components = (
    hibernate       => "AdvPowerMgr::Plugin::PreWakeup::Hibernate",

    poweron         => "AdvPowerMgr::Plugin::Wakeup::Program",

    delayedshutdown => "AdvPowerMgr::Plugin::PreShutdown::DelayedShutdown",
    suspend         => "AdvPowerMgr::Plugin::PreShutdown::Suspend",

    poweroff        => "AdvPowerMgr::Plugin::Shutdown::Program",

    networktraffic  => "AdvPowerMgr::Plugin::PostShutdown::Activity::Network",
    disktraffic     => "AdvPowerMgr::Plugin::PostShutdown::Activity::Disk",
    connections     => "AdvPowerMgr::Plugin::PostShutdown::Connections",
    hosts           => "AdvPowerMgr::Plugin::PostShutdown::Hosts",
    smart           => "AdvPowerMgr::Plugin::PostShutdown::SMART"
);

__PACKAGE__->_mk_accessors(qw( filename ));

foreach my $v (values(%components)) {
    eval "require $v"; die if $@;
}
__PACKAGE__->_mk_accessors(keys(%components));

sub prewakeup_plugins {
    my $self = shift;

    return ( $self->hibernate() );
}

sub wakeup_plugin {
    my $self = shift;

    return ( $self->poweron() );
}

sub preshutdown_plugins {
    my $self = shift;

    return ( $self->delayedshutdown(), $self->suspend() );
}

sub shutdown_plugin {
    my $self = shift;

    return ( $self->poweroff() );
}

sub postshutdown_plugins {
    my $self = shift;

    return ( $self->networktraffic(), $self->disktraffic(), $self->connections(), $self->hosts(), $self->smart() );
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params->{filename} = $class->_default_config_file() unless defined($params->{filename});

    if (defined($params->{objtype}) && $params->{objtype} ne $class) {
        die_ie(sprintf("unexpected objtype '%s'", $params->{objtype}));
    }

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->filename($params->{filename});

    while (my ($k, $v) = each(%components)) {
        $self->$k($v->new({ %{$params->{$k} || {}}, name => $k }));
    }

    return $self;
}

sub load {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params->{filename} = $class->_default_config_file() unless defined($params->{filename});
    $params->{use_defaults_on_failure} = 0 unless defined($params->{use_defaults_on_failure});

    my $file = $params->{filename};
    my $self;
    eval {
        if (open(my $fh, '<', $file)) {
            my @lines = (<$fh>);
            close($fh);
            my $obj;
            eval {
                $obj = from_json(join('', @lines));
            };
            if ($@) {
                info("Converting configuration");
                $self = $class->_convert_from_kvpairs($file, @lines);
            } else {
                $self = $class->new({ %$params, %$obj });
            }
        }
    };
    die $@ if $@ && !$params->{use_defaults_on_failure};

    unless ($self) {
        info("Using default configuration (fallback)");
        $self = $class->new({ filename => $file });
    }

    return $self;
}

sub save {
    my $self = shift;
    my $file = shift || $self->filename();

    my $json = to_json($self->to_hash({ with_class => 1 }));
    open(my $fh, '>', $file) or die_create($file);
    print $fh "$json";
    close($fh);

    return $self;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);

    foreach my $c (keys(%components)) {
        next if $c eq "suspend";
        $ret->{$c} = $self->$c()->to_hash($options);
    }

    return $ret;
}

sub _convert_from_kvpairs {
    my $class = shift;
    my $file = shift or die_param("file");

    my $self = $class->new({ filename => $file });

    my $wait_for_inactivity;
    my @poweron_easy = ();
    my @poweroff_easy = ();
    my @poweron_expert = ();
    my @poweroff_expert = ();
    my $mode;
    my %map = (
        mode => sub {
            my $v = shift;
            $mode = $v;
        },
        exp_on => sub {
            my $v = shift;
            push(@poweron_expert, $v);
        },
        exp_off => sub {
            my $v = shift;
            push(@poweroff_expert, $v);
        },
        wait_for_inactivity => sub {
            my $v = shift;
            if ($v eq "on") {
                $self->networktraffic()->enabled(1);
                $self->disktraffic()->enabled(1);
                $wait_for_inactivity = 1;
            } elsif ($v eq "off") {
                $self->networktraffic()->enabled(0);
                $self->disktraffic()->enabled(0);
                $wait_for_inactivity = 0;
            } else {
                die_ie("'$v' not supported (key 'wait_for_inactivity')");
            }
        },
        inactivity_interval => sub {
            my $v = shift;
            if ($v) {
                $self->networktraffic()->interval_min($v);
                $self->networktraffic()->checks_per_interval($v);
            }
        },
        network_threshold => sub {
            my $v = shift;
            if ($v) {
                $self->networktraffic()->enabled(1) if defined($wait_for_inactivity) && $wait_for_inactivity;
                $self->networktraffic()->threshold_mb($v);
            } else {
                $self->networktraffic()->enabled(0);
            }
        },
        disk_threshold => sub {
            my $v = shift;
            if ($v) {
                $self->disktraffic()->enabled(1) if defined($wait_for_inactivity) && $wait_for_inactivity;
                $self->disktraffic()->threshold_mb($v);
            } else {
                $self->disktraffic()->enabled(0);
            }
        },
        connections => sub {
            my $v = shift;
            if ($v eq "on") {
                $self->connections()->enabled(1);
            } elsif ($v eq "off") {
                $self->connections()->enabled(0);
            } else {
                die_ie("'$v' not supported (key 'connections')");
            }
        },
        lan_connections => sub {
            my $v = shift;
            if ($v eq "on") {
                $self->connections()->lan_only(1);
            } elsif ($v eq "off") {
                $self->connections()->lan_only(0);
            } else {
                die_ie("'$v' not supported (key 'lan_connections')");
            }
        },
        hibernation_date => sub {
            my $v = shift;
            $v =~ s#/#-#g;
            $self->hibernate()->date($v);
        },
        hibernation => sub {
            my $v = shift;
            if ($v eq "on") {
                $self->hibernate()->enabled(1);
            } elsif ($v eq "off") {
                $self->hibernate()->enabled(0);
            } else {
                die_ie("'$v' not supported (key 'hibernation')");
            }
        }
    );
    foreach my $day (qw( mon tue wed thr fri sat sun )) {
        $map{$day . "_off"} = sub { push(@poweroff_easy, "$day at $_[0]"); };
        $map{$day . "_on"} = sub { push(@poweron_easy, "$day at $_[0]"); };
    }

    my $kv_regex = qr/^(\w+)=(.+)$/;
    foreach (@_) {
        chomp();
        next unless $_ =~ $kv_regex;
        die_ie("unexpected key '$1'") unless exists($map{$1});
        $map{$1}->($2);
    }
    if ($mode eq "off") {
        $self->poweron()->enabled(0);
        $self->poweroff()->enabled(0);
        if (@poweron_expert || @poweroff_expert) {
            $self->poweron()->list(\@poweron_expert);
            $self->poweroff()->list(\@poweroff_expert);
        } else {
            $self->poweron()->list(\@poweron_easy);
            $self->poweroff()->list(\@poweroff_easy);
        }
    } else {
        $self->poweron()->enabled(1);
        $self->poweroff()->enabled(1);
        if ($mode eq "easy") {
            $self->poweron()->list(\@poweron_easy);
            $self->poweroff()->list(\@poweroff_easy);
        } else {
            $self->poweron()->list(\@poweron_expert);
            $self->poweroff()->list(\@poweroff_expert);
        }
    }

    return $self;
}

{
    our $_default_config_file;

    sub _default_config_file {
        my $class = shift;

        unless (defined($_default_config_file)) {
            my $dir = sprintf("%s/config", AdvPowerMgr::System->storedir());
            mkpath($dir);
            $_default_config_file = sprintf("%s/default", $dir);
        }

        return $_default_config_file;
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
