# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Daemon;

use strict;
use warnings;

use base qw ( AdvPowerMgr::Class );

use File::Basename qw( basename );
use POSIX qw();

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::System;

use constant KILL_TIMEOUT => 5;

our $name = basename($0);
our $pidfile = sprintf("%s/%s.pid", AdvPowerMgr::System->tmpdir(), $name);

sub kill {
    my $class = shift;

    my $fh;
    unless (open($fh, '<', $pidfile)) {
        info("No daemon '%s' is currently running", $name);
        return;
    }
    my $pid = <$fh>;
    close($fh);
    info("Daemon '%s' has PID %d", $name, $pid);

    my $p = CORE::kill('TERM', $pid);
    my $end = time() + KILL_TIMEOUT;
    while ((-f $pidfile) && (time() < $end)) {
            sleep(1);
    }
    if (-f $pidfile) {
        warning("Daemon '%s' is still alive!", $name);
        CORE::kill('KILL', $pid);
        unlink($pidfile);
    }
    info("Killed daemon '%s'", $name);
}

sub exists {
    my $class = shift;

    my $fh;
    unless (open($fh, '<', $pidfile)) {
        info("No daemon '%s' is currently running", $name);
        return 0;
    }
    my $pid = <$fh>;
    close($fh);
    info("Daemon '%s' has PID %d", $name, $pid);

    if (! CORE::kill('ZERO', $pid)) {
        warning("No daemon '%s' is currently running even though a pidfile exists!", $name);
        return -1;
    }

    return 1;
}

sub daemonize {
    my $class = shift;
    my $method = shift or die_param("method");

    my $pid = fork();
    if ($pid != 0) {
        info("Daemon '%s' has PID %d", $name, $pid);
        if (! open(my $f, '>', $pidfile)) {
            CORE::kill('KILL', $pid);
            die_create($pidfile);
        } else {
            print $f "$pid";
            close($f);
        }
        return;
    }

    open(STDIN, '<', '/dev/null');
    open(STDOUT, '>', '/dev/null');
    open(STDERR, '>', '/dev/null');
    chdir('/');
    POSIX::setsid();

    $SIG{INT} = $SIG{TERM} = sub {
        unlink($pidfile);
        info("Daemon '%s' exited", $name);
        exit(0);
    };
    $SIG{PIPE} = $SIG{HUP} = 'ignore';

    eval {
        $method->();
    };
    if ($@) {
        unlink($pidfile);
        my $msg = sprintf("Daemon '%s' exited abnormally: %s", $name, $@);
        fatal($msg);
        AdvPowerMgr::System->notify(i18n('internal_error.msg', $msg));
        eval "require AdvPowerMgr::Home;";
        unless ($@) {
            eval {
                AdvPowerMgr::Home->post_report($msg);
            };
            fatal($@) if $@;
        }
        exit(2);
    } else {
        unlink($pidfile);
        info("Daemon '%s' exited gracefully", $name);
        exit(0);
    }
}

sub exec {
    my $class = shift;
    my $cmd = shift or die_param("cmd");

    unlink($pidfile) if -e $pidfile;
    exec($cmd) or die_exec($cmd);
    exit(0);
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
