# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::HTTP;

use strict;
use warnings;

use base qw ( AdvPowerMgr::Class );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;

__PACKAGE__->_mk_accessors(qw( method data params ));

sub parse {
    my $class = shift;
    my $options = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    foreach my $var (qw( REQUEST_METHOD )) {
        die_param($var) unless defined($ENV{$var});
    }

    my $method = uc($ENV{REQUEST_METHOD});
    debug("Received '%s' HTTP request", $method);
    return undef if $method =~ m/^(OPTIONS|HEAD)$/;
    die "'$method' not allowed" if defined($options->{allowed_methods}) && !grep(/^$method$/i, @{$options->{allowed_methods}});

    my $self = $class->new();
    $self->method($method);

    if (defined($options->{allowed_params})) {
        if (ref($options->{allowed_params}) eq "ARRAY") {
            $options->{allowed_params} = { $method => $options->{allowed_params} };
        }
    } else {
        $options->{allowed_params} = { $method => [ ".*" ] };
    }
    my $ap = join('|', @{$options->{allowed_params}->{$method}});

    my $allowed_regex = qr/^(?:$ap)$/;
    my $skip_regex = qr/^(?:_dc|SynoToken)$/;

    my $params = {};
    if (defined($ENV{QUERY_STRING})) {
        foreach (split(/&/, $ENV{QUERY_STRING})) {
            my ($k, $v) = split(/=/, $_, 2);
            next if $k =~ $skip_regex;
            die "'$k' not allowed" unless $k =~ $allowed_regex;
            info("Request parameter '%s'='%s' (raw)", $k, $v);
            $v = $class->decode_urlencoded($v);
            $v =~ s/<!--(.|\n)*-->//g;

            $params->{$k} = $v;
        }
    }
    $self->params($params);

    if ($method ne "GET") {
        foreach my $var (qw( CONTENT_LENGTH CONTENT_TYPE )) {
            die_param($var) unless defined($ENV{$var});
        }

        die "Invalid content length" if defined($options->{allowed_data_length}) && $ENV{CONTENT_LENGTH} > $options->{allowed_data_length};
        my $data = "";
        if ($ENV{CONTENT_LENGTH} != 0) {
            if (defined($options->{stdin})) {
                $data = $options->{stdin};
            } else {
                read(STDIN, $data, $ENV{CONTENT_LENGTH});
            }
            my $ct = $ENV{CONTENT_TYPE};
            if ($ct =~ m#application/x-www-form-urlencoded#) {
                my $splitted = {};
                foreach (split(/&/, $data)) {
                    my ($k, $v) = split(/=/, $_);
                    $splitted->{$k} = $class->decode_urlencoded($v);
                }
                $data = $splitted;
            } else {
                die "Unexpected CONTENT_TYPE '$ct'";
            }
        }
        $self->data($data);
    }

    return $self;
}

sub decode_urlencoded {
    my $class = shift;
    my $data = shift;

    $data =~ tr/+/ /;
    $data =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

    return $data;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
