# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Home;

use strict;
use warnings;

use LWP::UserAgent;

use JSON qw( to_json );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;

{
    our $url = "http://reno.bozos.org:8080";
    our $ua = LWP::UserAgent->new(agent => "AdvPowerMgr", timeout => 10);

    sub post_enabled {
        my $class = shift;

        return ! -e sprintf("%s/nohome", AdvPowerMgr::System->storedir());
    }

    sub get_donators {
        my $class = shift;

        my @http = defined($ENV{HTTP_ACCEPT_LANGUAGE}) ? split(/,/, $ENV{HTTP_ACCEPT_LANGUAGE}) : ();
        my $l = AdvPowerMgr::System->lang();
        push(@http, "$l;q=1") if defined($l);
        push(@http, "en;q=0.001"); # Fallback
        $ua->timeout(120);
        my $response = $ua->post(sprintf("%s/cgi-bin/donators.pl", $url),
            {
                version => AdvPowerMgr::System->appversion(),
                hwid => AdvPowerMgr::System->hwid(),
                http_accept_language => join(',', @http)
            }
        );
        die $response->status_line unless $response->is_success;

        return $response->content();
    }

    sub post_lang {
        my $class = shift;
        my $lang = shift or die_param("lang");
        my $json = shift or die_param("json");

        unless ($class->post_enabled()) {
            warning("Not posting lang '%s' because posting is disabled", $lang);
            return;
        }

        my $response = $ua->post(sprintf("%s/cgi-bin/i18n.pl", $url),
            {
                lang => $lang,
                version => AdvPowerMgr::System->appversion(),
                hwid => AdvPowerMgr::System->hwid(),
                json => $json
            }
        );
        die $response->status_line unless $response->is_success;

        return $response->content();
    }

    sub post_generic {
        my $class = shift;
        my $what = shift or die_param("what");
        my $content = shift or die_param("content");

        unless ($class->post_enabled()) {
            warning("Not posting '%s' because posting is disabled", $what);
            return;
        }

        my $response = $ua->post(sprintf("%s/cgi-bin/generic.pl", $url),
            {
                what => $what,
                content => $content
            }
        );
        die $response->status_line unless $response->is_success;

        return $response->content();
    }

    sub post_report {
        my $class = shift;
        my $error = shift or die_param("error");
        my $config = shift;
        my $log = shift || AdvPowerMgr::Log->default();

        unless ($class->post_enabled()) {
            warning("Not posting report because posting is disabled");
            return;
        }

        if (defined($config)) {
            $config = $config->to_hash({ for_json => 1 });
        } else {
            $config = {};
        }

        $log = $log->dump();

        my $response = $ua->post(sprintf("%s/cgi-bin/report.pl", $url),
            {
                version => AdvPowerMgr::System->appversion(),
                hwid => AdvPowerMgr::System->hwid(),
                error => $error,
                config => to_json($config),
                log => $log
            }
        );
        die $response->status_line unless $response->is_success;

        return $response->content();
    }

    sub init {
        my $class = shift;
        my $proxy = AdvPowerMgr::System->proxy();

        return unless defined($proxy);
        if (defined($proxy->{user})) {
            $ua->proxy('http', sprintf("http://%s:%s@%s:%d",
                    $proxy->{user}, $proxy->{passwd}, $proxy->{host}, $proxy->{port}));
        } else {
            $ua->proxy('http', sprintf("http://%s:%d",
                    $proxy->{host}, $proxy->{port}));
        }
        return;
    }

}

__PACKAGE__->init();

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
