# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::I18N;

use strict;
use warnings;

use base qw ( AdvPowerMgr::Class );

use File::Path qw( mkpath );

use HTTP::AcceptLanguage;
use JSON qw( from_json );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;

our @EXPORT_OK;
push(@EXPORT_OK, qw( i18n ));

our $I18N_BUILTIN_DIR = sprintf("%s/i18n", AdvPowerMgr::System->appdir());
our $I18N_USER_DIR = sprintf("%s/i18n", AdvPowerMgr::System->storedir());

sub get {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    my @catalogs = ();
    if (defined($params->{catalog})) {
        push(@catalogs, $params->{catalog});
    } else {
        push(@catalogs, 'user', 'builtin');
    }

    my $pattern = defined($params->{pattern}) ? $params->{pattern} : qr/^.*$/;

    my %msg = ();
    my @used_langs = ();
    my %used_langs = ();
    my $count = 0;

    if (defined($params->{langs})) {
        foreach my $lang (@{$params->{langs}}) {
            foreach my $cat (@catalogs) {
                my $r = $class->_getfile($lang, $cat);
                while (my ($k, $v) = each(%$r)) {
                    next if exists($msg{$k});
                    $count++;
                    next unless $k =~ $pattern;
                    push(@used_langs, $lang) unless $used_langs{$lang}++;
                    $msg{$k} = $v;
                }
                goto DONE if $count == $class->_msgcount();
            }
        }
    } else {
        foreach my $cat (@catalogs) {
            my %list = ();
            foreach ($class->list({ catalog => $cat })) {
                $list{$_} = 1;
            }
            while (my $lang = $class->bestlang(keys(%list))) {
                my $r = $class->_getfile($lang, $cat);
                while (my ($k, $v) = each(%$r)) {
                    next if exists($msg{$k});
                    $count++;
                    next unless $k =~ $pattern;
                    push(@used_langs, $lang) unless $used_langs{$lang}++;
                    $msg{$k} = $v;
                }
                goto DONE if $count == $class->_msgcount();
                delete $list{$lang};
            }
        }
    }

DONE:
    info("Languages ['%s'] are used", join("', '", @used_langs));

    return { msgtable => \%msg, langs => \@used_langs };
}

{
    our $lang_cache;

    sub reset {
        my $class = shift;
        undef($lang_cache);
    }

    sub set {
        my $class = shift;
        my $lang = shift or die_param("lang");
        my $data = shift or die_param("data");

        $class->_setfile($lang, $data);
        $class->reset($lang);
    }

    sub i18n {
        my $class = shift if $_[0] eq __PACKAGE__; $class = __PACKAGE__ unless defined($class);
        my $msgid = shift;

        unless (defined($lang_cache)) {
            $lang_cache = {};

            foreach my $cat (qw( user builtin )) {
                my %list = ();
                foreach ($class->list({ catalog => $cat })) {
                    $list{$_} = 1;
                }
                while (my $lang = $class->bestlang(keys(%list))) {
                    my $r = $class->_getfile($lang, $cat);
                    while (my ($k, $v) = each(%$r)) {
                        $lang_cache->{$k} = $v unless exists($lang_cache->{$k});
                    }
                    goto DONE if (keys(%{$lang_cache}) == $class->_msgcount());
                    debug("Message catalog '%s' for language '%s' is incomplete", $cat, $lang);
                    delete $list{$lang};
                }
            }
        }

    DONE:
        die_ie("no msgid '$msgid'") unless exists($lang_cache->{$msgid});

        return sprintf($lang_cache->{$msgid}, @_);
    }
}

{
    our $list_regex = qr/^(.*)\.json$/;

    sub list {
        my $class = shift;
        my $options = shift || {};

        my @dirs;
        if (defined($options->{catalog})) {
            if ($options->{catalog} eq 'builtin') {
                @dirs = ( $I18N_BUILTIN_DIR );
            } else {
                @dirs = ( $I18N_USER_DIR );
            }
        } else {
            @dirs = ( $I18N_BUILTIN_DIR, $I18N_USER_DIR );
        }
        my %ret = ();

        foreach my $dir (@dirs) {
            opendir(my $dh, $dir) or next;
            while (my $f = readdir($dh)) {
                next unless $f =~ $list_regex;
                $ret{$1} = 1;
            }
            closedir($dh);
        }

        return sort(keys(%ret));
    }
}

sub _getfile {
    my $class = shift;
    my $lang = shift or die_param("lang");
    my $catalog = shift or die_param("catalog");

    my $ret = {};
    my $filename = sprintf("%s/%s.json", $catalog eq 'builtin' ? $I18N_BUILTIN_DIR : $I18N_USER_DIR, $lang);
    if (open(my $fh, '<', $filename)) {
        my @content = (<$fh>);
        close($fh);
        $ret = from_json(join('', @content));
    } elsif (! $!{ENOENT}) {
        die_open($filename);
    } else {
        debug("No language file '%s' for catalog '%s'", $lang, $catalog);
    }

    return $ret;
}

sub _setfile {
    my $class = shift;
    my $lang = shift or die_param("lang");
    my $data = shift or die_param("data");

    mkpath($I18N_USER_DIR);
    my $filename = sprintf("%s/%s.json", $I18N_USER_DIR, $lang);
    my $json = JSON->new()->canonical->pretty->encode($data);
    if (open(my $fh, '>', $filename)) {
        print $fh "$json";
        close($fh);
    } else {
        die_create($filename);
    }
}

{
    our $_msgcount;
    sub _msgcount {
        my $class = shift;

        unless ($_msgcount) {
            $_msgcount = keys(%{$class->_getfile('en', 'builtin')});
            info("There are %d messages in builtin catalog", $_msgcount);
        }

        return $_msgcount;
    }
}

{
    our $bestlang_al;

    sub bestlang {
        my $class = shift;
        my @available = @_;

        unless (defined($bestlang_al)) {
            my @http = defined($ENV{HTTP_ACCEPT_LANGUAGE}) ? split(/,/, $ENV{HTTP_ACCEPT_LANGUAGE}) : ();
            my $l = AdvPowerMgr::System->lang();
            unshift(@http, "$l;q=1") if defined($l); # Put system lang first to override browser's preferred one
            push(@http, "en;q=0.001"); # Fallback
            $bestlang_al = HTTP::AcceptLanguage->new(join(',', @http));
        }

        return $bestlang_al->match(@available);
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
