# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Log;

use strict;
use warnings;

use base qw( AdvPowerMgr::Class );

use File::Basename qw( basename );
use File::Path qw( mkpath );
use Scalar::Util qw( blessed );
use Symbol qw( qualify_to_ref );

use JSON qw( );

use AdvPowerMgr::Errors;

__PACKAGE__->_mk_accessors(qw( level with_time with_context with_package ));

use constant {
    FATAL       => 0,
    ERROR       => 1,
    WARN        => 2,
    INFO        => 3,
    DEBUG       => 4,
    DEBUG_FUNC  => 5,
    STR => {
        0 => 'FATAL',
        1 => 'ERROR',
        2 => 'WARN',
        3 => 'INFO',
        4 => 'DEBUG',
        5 => 'DEBUG_FUNC'
    },
};

use constant {
    DEFAULTS    => {
        level           => DEBUG_FUNC,
        with_context    => 1,
        with_package    => 1,
        with_time       => 1
    }
};

our @EXPORT_OK;
push(@EXPORT_OK, qw( DEBUG INFO WARN ERROR FATAL debug info warning error fatal debug_func_enter ));

our @EXPORT;
push(@EXPORT, qw( debug info warning error fatal debug_func_enter ));

__PACKAGE__->_mk_accessors(qw( default context ));

__PACKAGE__->context(basename($0));

sub levelstr {
    my $class = shift;
    my $level = shift; die_param('level') unless defined($level);

    return STR->{$level} or die_ie('level');
}

{
    our $maxlevel;

    sub maxlevel {
        my $class = shift;
        unless (defined($maxlevel)) {
            eval "require AdvPowerMgr::System;"; die if $@;
            my $file = sprintf("%s/loglevel", AdvPowerMgr::System->storedir());
            if (open(my $fh, '<', $file)) {
                my $val = <$fh>;
                chomp($val);
                close($fh);
                my $found = 0;
                foreach my $k (keys(%{(STR)})) {
                    if (STR->{$k} eq $val) {
                        $maxlevel = $k;
                        $found = 1;
                        last;
                    }
                }
                unless ($found) {
                    warning("Unexpected content '%s' in file '%s'", $val, $file);
                    $maxlevel = DEBUG_FUNC;
                }
            } else {
                $maxlevel = DEBUG_FUNC;
            }
        }
        return $maxlevel;
    }
}

sub default_trigger {
    my $class = shift;
    my $was = shift;

    # Auto-merge bootstrap logger to default logger
    return unless defined($was) && $was->isa("AdvPowerMgr::Log::Memory") && $was->name() eq "Bootstrap";
    my $default = $class->default();
    foreach ($was->dump()) {
        $default->_log($_) unless $_->{level} > $default->level();
    }
}

{
    our $defaults = DEFAULTS;
    sub defaults {
        my $class = shift;

        if (@_) {
            # Setter
            my $val = shift;
            $defaults = { %{(DEFAULTS)}, %$val };
            return $class;
        } else {
            # Getter
            return %$defaults;
        }
    }
}

sub _new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params = { $class->defaults(), %$params };

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->level($params->{level});
    $self->with_package($params->{with_package});
    $self->with_context($params->{with_context});
    $self->with_time($params->{with_time});

    return $self;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{level} = $self->level();
    $ret->{with_time} = $self->with_time();
    $ret->{with_package} = $self->with_package();
    $ret->{with_context} = $self->with_context();

    return $ret;
}

{
    our $debug_func_json;

    sub debug_func_enter {
        my $self = $_[0];
        if (defined($self) && blessed($self) && $self->isa(__PACKAGE__)) {
            shift;
        } else {
            shift if defined($self) && $self eq __PACKAGE__;
            $self = __PACKAGE__->default();
            unless (defined($self)) {
                eval "require AdvPowerMgr::Log::Memory;"; die if $@;
                $self = AdvPowerMgr::Log::Memory->new({ name => "Bootstrap" });
                __PACKAGE__->default($self);
            }
        }
        return if $self->level() < DEBUG_FUNC;
        my $hash = shift || {};
        unless (defined($debug_func_json)) {
            $debug_func_json = JSON->new->allow_nonref->allow_unknown->allow_blessed->convert_blessed;
        }
        my $time = time();
        my $package = caller(0);
        my $func = [caller(1)]->[3];
        my $msg = sprintf("%s(%s)", $func, $debug_func_json->encode($hash));
        $self->_log({ level => DEBUG, context => __PACKAGE__->context(), package => $package, time => $time, msg => $msg });
    }
}

sub _mk_method {
    my $class = shift;
    my $method = shift;
    my $level = shift;

    no strict 'refs';
    my $ref = qualify_to_ref($method, $class);

    *{$ref} = sub {
        my $self = $_[0];
        if (defined($self) && blessed($self) && $self->isa(__PACKAGE__)) {
            shift;
        } else {
            shift if defined($self) && $self eq __PACKAGE__;
            $self = __PACKAGE__->default();
            unless (defined($self)) {
                eval "require AdvPowerMgr::Log::Memory;"; CORE::die if $@;
                $self = AdvPowerMgr::Log::Memory->new({ name => "Bootstrap" });
                __PACKAGE__->default($self);
            }
        }
        return unless $level <= $self->level();
        my $time = time();
        my $package = caller(0);
        my $msg = shift;
        $msg = sprintf($msg, @_) if @_;
        $self->_log({ level => $level, context => __PACKAGE__->context(), package => $package, time => $time, msg => $msg });
    };
}

sub clear {
    my $self = shift;
    return $self;
}

sub dump {
    my $self = shift;
    return ();
}

sub sync {
    my $self = shift;
    return $self;
}

__PACKAGE__->_mk_method('fatal', FATAL);
__PACKAGE__->_mk_method('error', ERROR);
__PACKAGE__->_mk_method('warning', WARN);
__PACKAGE__->_mk_method('info', INFO);
__PACKAGE__->_mk_method('debug', DEBUG);

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
