# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Log::File;

use strict;
use warnings;

use base qw( AdvPowerMgr::Log );

use File::Basename qw( dirname );
use File::Path qw( mkpath );
use Time::Local qw( timelocal );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Validators qw( strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( filename maxdays maxlines with_header ));
__PACKAGE__->_mk_accessors(qw( _fh _arr _fmt _header_printed ));

sub maxlines_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub maxdays_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub with_package_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub with_time_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub with_context_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub with_header_trigger {
    my $self = shift;
    $self->_header_printed(0);
}

use constant {
    HEADER_FMT          => "--- 8< -- %04d-%02d-%02d %02d:%02d:%02d -",
    HEADER_REGEX        => qr/^--- 8< -- (\d+)-(\d+)-(\d+) /,
    LINE_REGEX          => qr/^(\d+):(\d+):(\d+) /,
    DEFAULTS    => {
        filename        => sprintf("%s/log", AdvPowerMgr::System->storedir()),
        maxdays         => 7,
        maxlines        => 1000,
        with_header     => 1
    }
};

{
    our $defaults;
    sub defaults {
        my $class = shift;

        if (@_) {
            # Setter
            my $val = shift;
            $defaults = { %{(DEFAULTS)}, $class->SUPER::defaults(), %$val };
            return $class;
        } else {
            # Getter
            unless (defined($defaults)) {
                $defaults = { %{(DEFAULTS)}, $class->SUPER::defaults() };
            }
            return %$defaults;
        }
    }
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params = { $class->defaults(), %$params, level => $class->maxlevel() };

    my $self = $class->SUPER::_new($params);
    bless($self, $class);
    $self->filename($params->{filename});
    $self->maxdays($params->{maxdays});
    $self->maxlines($params->{maxlines});
    $self->with_header($params->{with_header});

    return $self;
}

sub _compute_fmt {
    my $self = shift;

    my @fmt = ();

    push(@fmt, "%02d:%02d:%02d") if $self->{with_time};
    push(@fmt, "%-5s"); # Level
    push(@fmt, "%-8s") if $self->{with_context};
    push(@fmt, "%-12s") if $self->{with_package};
    push(@fmt, "%s"); # Message

    $self->_fmt(join(" | ", @fmt));
}

sub _log {
    my $self = shift;
    my $params = shift or CORE::die('params');

    my @lines = ();

    if ($self->with_header() && !$self->_header_printed()) {
        my ($s, $m, $h, $dm, $mo, $y) = localtime($params->{time});
        my $header = sprintf(HEADER_FMT, $y+1900, $mo+1, $dm, $h, $m, $s);
        push(@lines, $header . "-" x (80 - length($header)));
        $self->_header_printed(1);
    }

    my @fmt_args = ();
    if ($self->with_time()) {
        my ($s, $m, $h) = localtime($params->{time});
        push(@fmt_args, $h, $m, $s);
    }
    push(@fmt_args, $self->levelstr($params->{level}));
    if ($self->with_context()) {
        push(@fmt_args, substr(AdvPowerMgr::Log->context(), 0, 8));
    }
    if ($self->with_package()) {
        my $package = $params->{package};
        s#AdvPowerMgr::## for $package;
        push(@fmt_args, substr($package, -12));
    }
    push(@fmt_args, $params->{msg});
    push(@lines, split(/\n/, sprintf($self->_fmt(), @fmt_args)));

    $self->_dolog(@lines);
}

sub _dolog {
    my $self = shift;
    my @lines = @_;

    my $fh = $self->_fh();
    unless (defined($fh)) {
        $self->_init();
        $fh = $self->_fh();
    }
    my $arr = $self->_arr();
    push(@$arr, @lines);

    if (@$arr <= $self->maxlines()) {
        print $fh join("\n", @lines);
        print $fh "\n" if @lines;
    } else {
        while (@$arr > $self->maxlines()) {
            my $l = shift(@$arr);
            if ($l =~ LINE_REGEX) {
                while ($arr->[0] !~ LINE_REGEX) {
                    shift(@$arr);
                }
            }
        }
        truncate($fh, 0);
        seek($fh, 0, 0);
        print $fh join("\n", @$arr);
        print $fh "\n" if @$arr;
    }
}

sub _init {
    my $self = shift;

    my $fh;
    my @arr = ();
    if (open($fh, '+<', $self->filename())) {
        @arr = (<$fh>);
        chomp(@arr);

        my ($s, $m, $h, $d, $mo, $y) = localtime(time() - $self->maxdays() * 86400);
        my $min_t = timelocal(0, 0, 0, $d, $mo, $y);

        for (my $i = 0; $i < @arr; $i++) {
            next unless $arr[$i] =~ HEADER_REGEX;
            my $t = timelocal(0, 0, 0, $3, $2 - 1, $1 - 1900);
            next if $t < $min_t;
            last if $i == 0;
            splice(@arr, 0, $i);

            truncate($fh, 0);
            seek($fh, 0, 0);
            print $fh join("\n", @arr);
            print $fh "\n" if @arr;
            last;
        }
    } elsif ($!{ENOENT}) {
        mkpath(dirname($self->filename()));
        open($fh, '>', $self->filename()) or CORE::die(sprintf("failed to create file '%s': %s\n", $self->filename(), $!));
    } else {
        CORE::die(sprintf("failed to open file '%s': %s\n", $self->filename(), $!));
    }

    # Autoflush
    my $oldfh = select($fh); $| = 1; select($oldfh);

    $self->_fh($fh);
    $self->_arr(\@arr);
}

sub DESTROY {
    my $self = shift;
    if (defined($self->_fh())) {
        close($self->_fh());
        $self->_fh(undef);
    }
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{filename} = $self->filename();
    $ret->{maxdays} = $self->maxdays();
    $ret->{maxlines} = $self->maxlines();
    $ret->{with_header} = $self->with_header();

    return $ret;
}

sub clear {
    my $self = shift;

    if ($self->_fh()) {
        close($self->_fh());
        $self->_fh(undef);
    }
    unlink($self->filename());
    $self->_arr([]);

    return $self;
}

sub dump {
    my $self = shift;

    return @{$self->_arr()};
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
