# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Log::FileHandle;

use strict;
use warnings;

use base qw( AdvPowerMgr::Log );

use Time::Local qw( timelocal );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Validators qw( strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( fh ));
__PACKAGE__->_mk_accessors(qw( _fmt ));

sub with_package_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub with_time_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub with_context_trigger {
    my $self = shift;
    $self->_compute_fmt();
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    die_param('fh') unless defined($params->{fh});
    $params = { $class->defaults(), %$params };

    my $self = $class->SUPER::_new($params);
    bless($self, $class);
    $self->fh($params->{fh});

    return $self;
}

sub _compute_fmt {
    my $self = shift;

    my @fmt = ();

    push(@fmt, "%02d:%02d:%02d") if $self->{with_time};
    push(@fmt, "%-5s"); # Level
    push(@fmt, "%-8s") if $self->{with_context};
    push(@fmt, "%-12s") if $self->{with_package};
    push(@fmt, "%s"); # Message

    $self->_fmt(join(" | ", @fmt));
}

sub _log {
    my $self = shift;
    my $params = shift or die_ie('params');

    my @fmt_args = ();
    if ($self->with_time()) {
        my ($s, $m, $h) = localtime($params->{time});
        push(@fmt_args, $h, $m, $s);
    }
    push(@fmt_args, $self->levelstr($params->{level}));
    if ($self->with_context()) {
        push(@fmt_args, substr(AdvPowerMgr::Log->context(), 0, 8));
    }
    if ($self->with_package()) {
        my $package = $params->{package};
        s#AdvPowerMgr::## for $package;
        push(@fmt_args, substr($package, -12));
    }
    push(@fmt_args, $params->{msg});
    my $fh = $self->fh();
    print $fh sprintf($self->_fmt(), @fmt_args) . "\n";
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
