# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PostShutdown;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin );

use JSON qw();

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( time2str );
use AdvPowerMgr::Validators qw( strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( checks_per_interval interval_min ));
__PACKAGE__->_mk_accessors(qw( _maxdrift _nextupdate _offset _values ));

sub min_check_s {
    return 10;
}

sub checks_per_interval_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub interval_min_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub checks_per_interval_trigger {
    my $self = shift;

    my $checks_per_interval = $self->checks_per_interval();
    my $interval_min = $self->interval_min() or return;

    my $min_check_s = $self->min_check_s();
    die i18n('plugin.invalid_interval', $min_check_s, 'checks_per_interval', 'interval_min')
        unless $interval_min * 60 / $checks_per_interval >= $min_check_s;

    $self->_reset();
}

sub interval_min_trigger {
    my $self = shift;

    my $checks_per_interval = $self->checks_per_interval() or return;
    my $interval_min = $self->interval_min();

    my $min_check_s = $self->min_check_s();
    die i18n('plugin.invalid_interval', $min_check_s, 'checks_per_interval', 'interval_min')
        unless $interval_min * 60 / $checks_per_interval >= $min_check_s;

    $self->_reset();
}

sub _maxdrift_trigger {
    my $self = shift;

    info("Max drift computed to %d seconds for plugin '%s'", $self->_maxdrift(), $self->name());
}

sub _nextupdate_trigger {
    my $self = shift;

    my $nextupdate = $self->_nextupdate();
    info("Plugin '%s' next check will be performed at '%s'", $self->name(), time2str($nextupdate));
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;
    my $options = shift || {};

    if (defined($params->{objtype}) && $params->{objtype} ne $class) {
        die_ie(sprintf("unexpected objtype '%s'", $params->{objtype}));
    }

    $params->{enabled} = 0 unless defined($params->{enabled});
    $params->{checks_per_interval} = 1 unless defined($params->{checks_per_interval});
    $params->{interval_min} = 1 unless defined($params->{interval_min});
    unless (defined($params->{name})) {
        $params->{name} = substr($class, rindex($class, "::") + 2);
    }

    my $self = $class->SUPER::new($params, { disable_trigger => 1 });
    bless($self, $class);
    $self->checks_per_interval($params->{checks_per_interval}, { disable_trigger => 1 });
    $self->interval_min($params->{interval_min}, { disable_trigger => 1 });
    $self->_reset() unless exists($options->{disable_trigger}) && $options->{disable_trigger};
    return $self;
}

sub offset {
    my $self = shift;

    return $self->_offset();
}

sub compute {
    my $self = shift;
    my $when = shift || time();

    return $self->_laststate() unless $self->enabled();

    if ($self->_nextupdate() == 0) {
        $self->_nextupdate($when, { disable_trigger => 1 });
    } elsif ($when < $self->_nextupdate()) {
        info("Plugin '%s' called too early", $self->name());
        return $self->_laststate();
    }

    if (($self->_maxdrift() != 0) && ($when > $self->_nextupdate() + $self->_maxdrift())) {
        warning("Time drifted too much for plugin '%s', resetting values", $self->name());
        $self->_reset();
        $self->_nextupdate($when, { disable_trigger => 1 });
    }

    my $newval = $self->_value_reader();

    # Case of asynchronous plugin, plugin must be called twice
    return undef unless defined($newval);

    my $poppedval = shift(@{$self->_values()});
    push(@{$self->_values()}, $newval);

    $self->_nextupdate($self->_nextupdate_compute($self->_nextupdate()));
    return $self->_compute_result($poppedval, $newval);
}

# Returns a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - 0 if no further update required
#   - timestamp otherwise
# - done (valid if enabled only)
#   - 0 if condition is not met
#   - 1 otherwise
# - checks_per_interval (static)
# - interval_min (static)
sub _compute_result {
    my $self = shift;
    my $poppedval = shift;      # Ignored (popped value)
    my $newval = shift;         # Ignored (pushed value)

    my $state = $self->_laststate();
    $state->{nextupdate_ts} = $self->_nextupdate();

    #
    # Examine all stored values
    #
    $state->{done} = 1;
    foreach my $val (@{$self->_values()}) {
        if (!defined($val) || $val != 0) {
            $state->{done} = 0;
            last; # foreach ($val)
        }
    }

    return $state;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    if ($options->{for_json}) {
        $ret->{enabled} = $self->enabled() ? JSON->true() : JSON->false();
    } else {
        $ret->{enabled} = $self->enabled();
    }
    $ret->{name} = $self->name();
    $ret->{checks_per_interval} = int($self->checks_per_interval());
    $ret->{interval_min} = int($self->interval_min());

    return $ret;
}

sub _value_reader;

sub _reset {
    my $self = shift;

    $self->SUPER::_reset();
    $self->_nextupdate(0, { disable_trigger => 1 });
    $self->_maxdrift(0, { disable_trigger => 1 });
    $self->_offset($self->enabled() ? $self->interval_min() * 60 : 0);
    $self->_values([ (undef) x $self->checks_per_interval() ]);

    if ($self->enabled()) {
        my $state = $self->_laststate();
        $state->{nextupdate_ts} = 0;
        $state->{checks_per_interval} = int($self->checks_per_interval());
        $state->{interval_min} = int($self->interval_min());
    }
}

sub _nextupdate_compute {
    my $self = shift;
    my $nextupdate = shift;

    return $nextupdate + 60 * $self->interval_min() / $self->checks_per_interval();
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
