# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PostShutdown::Activity;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown );

use AdvPowerMgr::Log;
use AdvPowerMgr::Validators qw( strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( threshold_mb ));

sub threshold_mb_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub threshold_mb_trigger {
    my $self = shift;

    $self->_reset();
}

use constant MB         => 1048576;

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;
    my $options = shift || {};

    debug_func_enter($params);

    $params->{threshold_mb} = 10 unless defined($params->{threshold_mb});

    if (defined($params->{objtype}) && $params->{objtype} =~ m/^AdvPowerMgr::Plugin::Interval::/) {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $self = $class->SUPER::new($params, { disable_trigger => 1 });
    bless($self, $class);
    $self->threshold_mb($params->{threshold_mb}, { disable_trigger => 1 });
    $self->_reset() unless exists($options->{disable_trigger}) && $options->{disable_trigger};
    return $self;
}

sub _value_comparator {
    my $self = shift;
    my $val1 = shift;
    my $val2 = shift;

    unless (defined($val1)) {
        return undef;
    }

    my $diff = $val2 - $val1;
    return undef if $diff < 0;
    return $diff;
}

# Returns a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - timestamp
# - done
#   - 0 if condition is not met
#   - 1 otherwise
# - checks_per_interval (static)
# - interval_min (static)
# - threshold_mb (static)
# - value_mb
sub _compute_result {
    my $self = shift;
    my $oldval = shift;
    my $newval = shift;

    my $diff = $newval - $oldval if defined($oldval) && ($newval - $oldval >= 0);

    my $state = $self->_laststate();
    $state->{nextupdate_ts} = $self->_nextupdate();
    $state->{value_mb} = $diff;

    if (!defined($diff)) {
        info("Activity not computable yet");
        $state->{done} = 0;
    } elsif ($diff >= $self->threshold_mb() * MB) {
        info("Activity detected (%0.2f MB/%d min)",
            $diff / MB, $self->interval_min());
        $state->{done} = 0;
    } else {
        info("No significant activity detected (%0.2f MB/%d min)",
            $diff / MB, $self->interval_min());
        $state->{done} = 1;
    }

    return $state;
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{threshold_mb} = int($self->threshold_mb());

    return $ret;
}

sub _reset {
    my $self = shift;

    $self->SUPER::_reset();
    $self->_maxdrift($self->interval_min() * 60 / $self->checks_per_interval() / 2);
    $self->_laststate()->{threshold_mb} = int($self->threshold_mb());
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
