# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PostShutdown::Connections;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown );

use JSON qw();

use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::System;
use AdvPowerMgr::Utils qw( ipv4_to_hex uniq );
use AdvPowerMgr::Validators qw( boolean_validator );

__PACKAGE__->_mk_accessors(qw( lan_only portlist ));
__PACKAGE__->_mk_accessors(qw( _ports ));

our $comments_regex = qr/^(?:([-#@].*)|\s*)$/;

sub lan_only_validator {
    my $self = shift;
    my $value = shift;

    return boolean_validator($value);
}

{
    our $trim_regex = qr/^\s+|\s+$/;

    sub portlist_validator {
        my $self = shift;
        my $value = shift;

        for (my $i = 0; $i < @$value; $i++) {
            next if $value->[$i] =~ $comments_regex;
            $value->[$i] =~ s/$trim_regex//g;

            die i18n('connections.invalid_port', $value->[$i]) unless $value->[$i] =~ m/^\d+$/ && $value->[$i] > 0 && $value->[$i] <= 65535;
        }
        return $value;
    }

    sub portlist_trigger {
        my $self = shift;

        my %ports = ();
        foreach my $port (@{$self->portlist()}) {
            next unless $port =~ m/^\d+$/;
            $ports{$port}++;
        }
        $self->_ports([ sort(keys(%ports)) ]);
    }
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;
    my $options = shift || {};

    debug_func_enter($params);

    $params->{lan_only} = 1 unless defined($params->{lan_only});

    die_ie("'ports' and 'portlist' are mutually exclusive") if defined($params->{ports}) && defined($params->{portlist});
    $params->{portlist} = [] unless defined($params->{portlist});
    $params->{portlist} = [ split(/\n/, $params->{ports}) ] if defined($params->{ports});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Connections") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $self = $class->SUPER::new($params, { disable_trigger => 1 });
    bless($self, $class);
    $self->lan_only($params->{lan_only});
    $self->portlist($params->{portlist});
    $self->_reset() unless exists($options->{disable_trigger}) && $options->{disable_trigger};
    return $self;
}

{
    our %networks;

    sub _value_reader {
        my $self = shift;

        my @conns = AdvPowerMgr::System->get_connections();
        my @hosts = ();

        if (@conns != 0) {
            my %ips = ();
            foreach my $c (@conns) {
                my ($r_ip, $l_port) = ($c->{remote_ip}, $c->{local_port});
                my $r_ip_hex = ipv4_to_hex($r_ip);
                my $r_ip_hexval = hex($r_ip_hex);

                if ($self->lan_only()) {
                    %networks = AdvPowerMgr::System->get_networks() unless %networks;
                    my $found = 0;
                    foreach my $net_hex (keys(%networks)) {
                        my $mask_hexval = hex($networks{$net_hex});
                        if (sprintf("%08x", $r_ip_hexval & $mask_hexval) eq $net_hex) {
                            $found = 1;
                            debug("Host '%s' is in LAN", $r_ip);
                            last;
                        }
                    }
                    if (!$found) {
                        debug("Host '%s' is not in LAN, ignoring", $r_ip);
                        next; # foreach my $c
                    }
                }
                if (@{$self->_ports()}) {
                    my $found = 0;
                    foreach my $port (@{$self->_ports()}) {
                        next unless $l_port == $port;
                        unless (exists($ips{$r_ip_hexval})) {
                            $ips{$r_ip_hexval} = { remote_ip => $r_ip, local_ports => [] };
                        }
                        $found = 1;
                        debug("Port %d is in list", $l_port);
                        last;
                    }
                    if (!$found) {
                        debug("Port %d is not in list, ignoring", $l_port);
                        next; # foreach my $c
                    }
                }

                unless (exists($ips{$r_ip_hexval})) {
                    $ips{$r_ip_hexval} = { remote_ip => $r_ip, local_ports => [] };
                }
                push(@{$ips{$r_ip_hexval}->{local_ports}}, int($l_port));
            }

            foreach my $ip_hex (sort(keys(%ips))) {
                $ips{$ip_hex}->{local_ports} = [ uniq(sort { $a <=> $b } @{$ips{$ip_hex}->{local_ports}}) ];
                push(@hosts, $ips{$ip_hex});
            }
        }
        my $count = 0;
        map { $count += @{$_->{local_ports}} } @hosts;
        info("There are %d connected ports", $count);
        $self->_laststate()->{count} = $count;
        $self->_laststate()->{hosts} = \@hosts;

        return $count;
    }

    sub _reset {
        my $self = shift;

        $self->SUPER::_reset();
        undef(%networks);
        if ($self->enabled()) {
            $self->_laststate()->{count} = 0;
            $self->_laststate()->{hosts} = [];
            $self->_laststate()->{lan_only} = $self->lan_only();
        }
    }
}

# compute() is expected to return a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - timestamp
# - done
#   - 0 if condition is not met
#   - 1 otherwise
# - checks_per_interval (static)
# - interval_min (static)
# - lan_only (static)
# - count
# - hosts

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    if ($options->{for_json}) {
        $ret->{lan_only} = $self->lan_only() ? JSON->true() : JSON->false();
    } else {
        $ret->{lan_only} = $self->lan_only();
    }
    $ret->{ports} = join("\n", @{$self->portlist()});

    return $ret;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
