# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PostShutdown::Hosts;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PostShutdown );

use Net::Ping;

use JSON qw();

use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::System;
use AdvPowerMgr::Validators qw( hostname_validator ipaddr_validator macaddr_validator strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( icmp_timeout syn_timeout list ));
__PACKAGE__->_mk_accessors(qw( _syn_ping _tmphosts ));

our $comments_regex = qr/^(?:([-#@].*)|\s*)$/;
our $full_regex = qr/^(\S+)(?:\s+port\s+(\d+))?$/i;

sub syn_timeout_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

sub icmp_timeout_validator {
    my $self = shift;
    my $value = shift;

    return strictly_positive_integer_validator($value);
}

{
    our $trim_regex = qr/^\s+|\s+$/;

    sub list_validator {
        my $self = shift;
        my $value = shift;

        for (my $i = 0; $i < @$value; $i++) {
            next if $value->[$i] =~ $comments_regex;
            $value->[$i] =~ s/$trim_regex//g;

            $value->[$i] =~ $full_regex or die i18n('hosts.invalid_line', $value->[$i]);
            my ($h, $p) = ($1, $2);

            if ($h =~ m/:/) {
                $h = macaddr_validator($h);
            } elsif ($h =~ m/^\d+\./) {
                $h = ipaddr_validator($h);
            } else {
                $h = hostname_validator($h);
            }
            if (defined($p)) {
                die i18n('hosts.invalid_port', $p) unless $p > 0 && $p <= 65535;
                $value->[$i] = "$h port $p";
            } else {
                $value->[$i] = $h;
            }
        }
        return $value;
    }
}

sub timeout_trigger {
    my $self = shift;
    $self->_reset();
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    debug_func_enter($params);

    die_ie("'hosts' and 'list' are mutually exclusive") if defined($params->{list}) && defined($params->{hosts});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Hosts") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    $params->{list} = [] unless defined($params->{list});
    $params->{list} = [ split(/\n/, $params->{hosts}) ] if defined($params->{hosts});
    $params->{syn_timeout} = 5 unless defined($params->{syn_timeout});
    $params->{icmp_timeout} = 1 unless defined($params->{icmp_timeout});

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->list($params->{list});
    $self->syn_timeout($params->{syn_timeout});
    $self->icmp_timeout($params->{icmp_timeout});
    return $self;
}

sub _value_reader {
    my $self = shift;

    # For SYN pings
    my $p = $self->_syn_ping();

    #
    # Plugin is asynchronous when using SYN pings
    #
    if (defined($p)) {
        my $hosts = $self->_tmphosts();
        while (my ($host, $rtt, $ip) = $p->ack()) {
            info("Host '%s' (%s) is alive (answered in %.2f ms)", $host, $ip, 1000 * $rtt);
            $hosts->{$host} = 1;
        }

        my $count = scalar(keys(%{$hosts}));
        info("There are %d hosts alive", $count);
        $self->_laststate()->{hosts} = [ sort(keys(%{$hosts})) ];
        $self->_syn_ping(undef);
        $self->_tmphosts({});

        return $count;
    }

    my @syn = ();
    my @icmp = ();
    my @macs = (); # Array of MAC addresses to be resolved

    foreach my $text (@{$self->list()}) {
        next if $text =~ $comments_regex;
        $text =~ $full_regex or die i18n('hosts.invalid_line', $text);
        my ($host, $port) = ($1, $2);
        if ($host =~ /:/) {
            push(@macs, [ $host, $port ]);
        } elsif (defined($port)) {
            push(@syn, [ $host, $port ]);
        } else {
            push(@icmp, $host);
        }
    }

    if (@macs) {
        my %arpcache = AdvPowerMgr::System->get_arp_cache();
        foreach my $entry (@macs) {
            my ($mac, $port) = @$entry;
            unless (exists($arpcache{$mac})) {
                warning("MAC address '%s' not in ARP cache, assuming host is down", $mac);
                next;
            }
            foreach my $host (@{$arpcache{$mac}}) {
                info("MAC address '%s' is '%s'", $mac, $host);
                if (defined($port)) {
                    push(@syn, [ $host, $port ]);
                } else {
                    push(@icmp, $host);
                }
            }
        }
    }

    my $hosts = {};

    #
    # Do the actual pings
    #

    if (@icmp) {
        my $icmp_p = Net::Ping->new("icmp", $self->icmp_timeout());
        $icmp_p->hires();
        foreach my $host (@icmp) {
            info("Ping '%s' using ICMP", $host);
            my ($res, $rtt, $ip) = $icmp_p->ping($host);
            if (!defined($res) || $res == 0) {
                info("Host '%s' is not reachable", $host);
            } else {
                info("Host '%s' (%s) is alive (answered in %.2f ms)", $host, $ip, 1000 * $rtt);
                $hosts->{$host} = 1;
            }
        }
    }

    if (@syn) {
        $self->_tmphosts($hosts);

        $p = Net::Ping->new("syn", $self->syn_timeout());
        $p->hires();
        $self->_syn_ping($p);

        foreach my $entry (@syn) {
            my ($host, $port) = @$entry;
            info("Pinging '%s' on port %d using SYN", $host, $port);
            $p->port_number($port);
            $p->ping($host);
        }

        return undef; # Asynchronous call (will be called twice)
    }

    my $count = scalar(keys(%{$hosts}));
    info("There are %d hosts alive", $count);
    $self->_laststate()->{hosts} = [ sort(keys(%{$hosts})) ];

    #
    # We are synchronous (only ICMP pings)
    #
    return $count;
}

# compute() is expected to return a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - timestamp
# - done
#   - 0 if condition is not met
#   - 1 otherwise
# - checks_per_interval (static)
# - interval_min (static)
# - hosts

sub _reset {
    my $self = shift;

    $self->SUPER::_reset();
    $self->_syn_ping(undef);
    if ($self->enabled()) {
        $self->_laststate()->{hosts} = [];
    }
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{syn_timeout} = int($self->syn_timeout());
    $ret->{icmp_timeout} = int($self->icmp_timeout());
    $ret->{hosts} = join("\n", @{$self->list()});

    return $ret;
}

unless (Net::Ping->can("port_number")) {
    #
    # Net::Ping->port_number() doesn't exist on Perl 5.8.6
    #
    eval "use Symbol qw( qualify_to_ref );"; die if $@;
    no strict 'refs';
    my $ref = qualify_to_ref("port_number", "Net::Ping");
    *{$ref} = sub {
        my $self = shift;
        my $port = shift; $port = 7 unless defined($port);
        $self->{port_num} = $port;
        $self->service_check(1);
        return $port;
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
