# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PreShutdown::DelayedShutdown;

#
# This module makes the shutdown phase occur after a given time since boot.
#

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PreShutdown );

use JSON qw();

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Validators qw( strictly_positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( delay_min ));

use constant MIN_DELAY_MIN  =>  10;

sub delay_min_validator {
    my $class = shift;
    my $value = shift;

    my $res = strictly_positive_integer_validator($value);
    die i18n('delayedshutdown.invalid_delay', MIN_DELAY_MIN) if $res < MIN_DELAY_MIN;
    return $res;
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;
    my $options = shift || {};

    debug_func_enter($params);

    $params->{delay_min} = 3 * MIN_DELAY_MIN unless defined($params->{delay_min});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::DelayedShutdown") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $self = $class->SUPER::new($params, { disable_trigger => 1 });
    bless($self, $class);
    $self->delay_min($params->{delay_min});
    $self->_reset() unless exists($options->{disable_trigger}) && $options->{disable_trigger};
    return $self;
}

# compute() is expected to return a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - 0
# - shutdown_ts
#   - 0 if no delayed shutdown planned
#   - timestamp otherwise
sub compute {
    my $self = shift;
    my $when = shift || time();

    if ($self->enabled()) {
        $self->_laststate()->{shutdown_ts} = int($when + $self->delay_min() * 60);
    }

    return $self->_laststate();
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{delay_min} = $self->delay_min();

    return $ret;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
