# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::PreShutdown::Suspend;

#
# This module makes the shutdown phase occur after a given date.
#

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin::PreShutdown );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;
use AdvPowerMgr::Validators qw( positive_integer_validator );

__PACKAGE__->_mk_accessors(qw( until ));

sub until_validator {
    my $class = shift;
    my $value = shift;

    return positive_integer_validator($value);
}

{
    our $file;

    sub _file {
        my $class = shift;

        unless (defined($file)) {
            $file = sprintf("%s/suspend", AdvPowerMgr::System->tmpdir());
        }
        return $file;
    }

    #
    # This module is special, it is not a real configuration plugin and is a singleton
    #
    our $singleton;

    sub new {
        my $class = shift;
        my $params = shift || {};
        my $options = shift || {};

        if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Suspend") {
            # Upgrade from 3.x, x <= 2
            delete $params->{objtype};
        }

        unless (defined($singleton)) {
            $singleton = $class->SUPER::new($params, { disable_trigger => 1 });
            bless($singleton, $class);
        }
        $singleton->_reset() unless exists($options->{disable_trigger}) && $options->{disable_trigger};
        return $singleton;
    }
}

sub suspend {
    my $self = shift;
    my $until = shift || -1;

    my $file = $self->_file();
    debug("Creating file '%s'", $file);
    open(my $fh, '>', $file) or die_create($file);
    print $fh "$until";
    close($fh);

    return $self->compute();
}

sub resume {
    my $self = shift;

    my $file = $self->_file();
    if (-e $file) {
        debug("Deleting file '%s'", $file);
        unlink($file);
    }

    return $self->compute();
}

sub to_hash {
    my $self = shift;

    my $ret = $self->SUPER::to_hash();
    delete $ret->{enabled};

    return $ret;
}

# compute() is expected to return a hash with the following fields:
# - enabled
#   - 0 if plugin is disabled
#   - 1 otherwise
# - nextupdate_ts
#   - 0 if suspended
#   - timestamp otherwise
# - shutdown_ts
#   - 0 if no suspend
#   - -1 if should never shut down
#   - timestamp otherwise
sub compute {
    my $self = shift;
    my $when = shift || time();

    $self->_reset($when);

    return $self->SUPER::compute($when);
}

sub _reset {
    my $self = shift;
    my $when = shift || time();

    my $file = $self->_file();
    $self->enabled(-e $file || 0, { disable_trigger => 1 });
    $self->SUPER::_reset();

    if ($self->enabled()) {
        open(my $fh, '<', $file) or die_open($file);
        my $until = <$fh>;
        close($fh);
        $until = -1 unless defined($until); # Handles case of upgrade from pre 3.2
        $self->_laststate()->{shutdown_ts} = int($until); # Get down at this time
        if ($until == -1 || $until <= $when) {
            $self->_laststate()->{nextupdate_ts} = 0;
        } else {
            $self->_laststate()->{nextupdate_ts} =  int($until);
        }
    } else  {
        $self->_laststate()->{nextupdate_ts} = 0;
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
