# Copyright (c) 2010-2015 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Plugin::Program;

use strict;
use warnings;

use base qw( AdvPowerMgr::Plugin );

use Time::Local qw( timelocal );

use AdvPowerMgr::Errors;
use AdvPowerMgr::I18N qw( i18n );
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( min_ts max_ts uniq );

__PACKAGE__->_mk_accessors(qw( list ));

sub list_validator {
    my $self = shift;
    my $value = shift;

    __PACKAGE__->_compute(0, $value);

    return [ __PACKAGE__->_normalize(@{$value}) ];
}

sub new {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    debug_func_enter($params);

    die_ie("'program' and 'list' are mutually exclusive") if defined($params->{list}) && defined($params->{program});

    $params->{list} = [] unless defined($params->{list});
    $params->{list} = [ split(/\n/, $params->{program}) ] if defined($params->{program});

    if (defined($params->{objtype}) && $params->{objtype} eq "AdvPowerMgr::Plugin::Program") {
        # Upgrade from 3.x, x <= 2
        delete $params->{objtype};
    }

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->list($params->{list});
    return $self;
}

{
    our @weekdays = ( "Sun", "Mon", "Tue", "Wed", "Thr", "Fri", "Sat" );
    our @months = ( "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" );

    our $comments_regex = qr/^([-#@].*)?$/;
    our $spaces_regex = qr/\s+/;
    our $trim_regex = qr/^\s+|\s+$/;
    our $range_regex = qr/\s-\s?/;
    our $list_regex = qr/\s?,\s?/;
    our $word_regex = qr/(\w+)/;
    our $word_mapper = sub {
        my $word = shift;
        my $lc = lc($word);
        return uc($word) if $lc eq "daily";
        return lc($word) if $lc eq "at" || $lc eq "except";
        return ucfirst($lc);
    };

    sub _normalize {
        my $class = shift;
        my @arr = @_;

        for (my $i = 0; $i < @arr; $i++) {
            next if $arr[$i] =~ $comments_regex; # don't normalize comments
            $arr[$i] =~ s/$trim_regex//g;
            $arr[$i] =~ s/$spaces_regex/ /g;
            $arr[$i] =~ s/$word_regex/$word_mapper->($1)/ge;
            $arr[$i] =~ s/$range_regex/-/g;
            $arr[$i] =~ s/$list_regex/, /g;
        }
        return @arr;
    }
}

sub to_hash {
    my $self = shift;
    my $options = shift || {};

    my $ret = $self->SUPER::to_hash($options);
    $ret->{program} = join("\n", @{$self->list()});

    return $ret;
}

sub compute {
    my $self = shift;
    my $when = shift || time();

    return 0 if !$self->enabled();

    return __PACKAGE__->_compute($when, $self->list());
}

sub _compute {
    my $class = shift;
    my $when = shift || time();
    my $list = shift or die_param("list");

    my @timetable = ();

    foreach my $maxdays_ahead (7, 365) {
        foreach my $line (@{$list}) {
            $line =~ s/\s+/ /g; $line =~ s/^ //; $line =~ s/ $//;
            next if $line =~ m/^([-#@].*)?$/; # Comments
            $line =~ s/ ,/,/g; $line =~ s/, /,/g;
            $line =~ s/ -/-/g; $line =~ s/- /-/g;
            $line = lc($line);
            $line =~ m/^(.+) at (.+) except (.+)$/i || $line =~ m/^(.+) at (.+)$/i or die i18n('program.invalid_line', $line);
            my ($datestr, $timestr, $except_datestr) = ($1, $2, $3);

            my @dates = _extract_dates($datestr, $when, $maxdays_ahead);
            if (defined($except_datestr)) {
                my @except_dates = _extract_dates($except_datestr, $when, $maxdays_ahead);
                my $lastindex = 0;
                foreach my $ed (@except_dates) {
                    while ($lastindex < $#dates) {
                        last if ($dates[$lastindex] > $ed);
                        if ($dates[$lastindex] == $ed) {
                            splice(@dates, $lastindex, 1);
                        } else {
                            $lastindex++;
                        }
                    }
                }
            }

            my @times = _extract_times($timestr);

            foreach my $d (@dates) {
                foreach my $t (@times) {
                    # Because of daylight saving, we cannot just add $d + $t
                    my (undef, undef, undef, $dm, $mo, $yo) = localtime($d);
                    my ($s, $m, $h) = ($t % 60, int(($t % 3600) / 60), int($t / 3600));
                    my $ts = timelocal($s, $m, $h, $dm, $mo, $yo);

                    next unless $ts > $when;
                    push(@timetable, $ts);
                }
            }
        }
        last if @timetable;
    }

    return 0 if !@timetable;

    @timetable = uniq(sort { $a <=> $b } @timetable);

    return shift(@timetable);
}

{
    our @weekdays = ( "sun", "mon", "tue", "wed", "thr", "fri", "sat" );
    our @months = ( "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec" );
    our %weekdaysmap = (); for (my $i = 0; $i < @weekdays; $i++) { $weekdaysmap{$weekdays[$i]} = $i; };
    our %monthsmap = (); for (my $i = 0; $i < @months; $i++) { $monthsmap{$months[$i]} = $i; };

    # list of "daily | mon-wed | jun 15 | jun 16-jul 24 | dec 29 2014 | dec 29 2014-jan 5 2015"
    our $full_regex = qr/^(daily|\w+(?:-\w+)?|\w+ \d+(?:-\w+ \d+)?|\w+ \d+ \d{4}(?:-\w+ \d+ \d{4})?)(?:,(.*))?$/;
    our $weekday_regex = qr/^(\w+)(?:-(\w+))?$/;
    our $date_regex = qr/^(\w+) (\d+)(?:-(\w+) (\d+))?$/;
    our $datewithyear_regex = qr/^(\w+) (\d+) (\d{4})(?:-(\w+) (\d+) (\d{4}))?$/;

    sub _extract_dates {
        my ($datestr, $start_when, $maxdays_ahead) = @_;

        my @ret = ();

        my ($startwhen_wd, $startwhen_yo);
        {
            my ($s, $m, $h, $dm, $mo, $yo, $wd) = localtime($start_when);
            $start_when = timelocal(0, 0, 0, $dm, $mo, $yo);
            ($startwhen_wd, $startwhen_yo) = ($wd, $yo);
        }

        my $rest = lc($datestr);
        die i18n('program.unexpected_date_string', $rest) if $rest eq "";

        while ($rest =~ $full_regex) {
            my $item = $1;
            $rest = defined($2) ? $2 : "";
            if ($item eq "daily") { $item = "mon-sun"; }
            if ($item =~ $weekday_regex) {
                # A day or range of days
                my ($fd, $ld) = ($1, defined($2) ? $2 : $1);

                die i18n('program.invalid_day', $fd) unless exists($weekdaysmap{$fd});
                die i18n('program.invalid_day', $ld) unless exists($weekdaysmap{$ld});
                my $fd_offset = $weekdaysmap{$fd};
                my $ld_offset = $weekdaysmap{$ld} + ($weekdaysmap{$ld} < $weekdaysmap{$fd} ? 7 : 0);

                #
                # Compute intersection between
                # [ start_when ; start_when+7 days ] and
                # [$fd;$ld] (+/- 1 week)
                # sup can be overriden by $maxdays_ahead
                # (very unusual case)
                #
                foreach my $i (-7, 0, $maxdays_ahead) {
                    my ($inf, $sup) = (
                        max_ts($start_when, $start_when + ($fd_offset - $startwhen_wd + $i) * 86400),
                        min_ts($start_when + $maxdays_ahead * 86400, $start_when + ($ld_offset - $startwhen_wd + $i) * 86400)
                    );
                    for (my $t = $inf; $t <= $sup; $t += 86400) {
                        push(@ret, $t);
                    }
                }
            } elsif ($item =~ $date_regex) {
                # A date or range of dates
                my ($fm, $fd, $lm, $ld);
                if (defined($3)) {
                    ($fm, $fd, $lm, $ld) = ($1, $2, $3, $4);
                } else {
                    ($fm, $fd, $lm, $ld) = ($1, $2, $1, $2);
                }
                die i18n('program.invalid_month', $fm) unless exists($monthsmap{$fm});
                die i18n('program.invalid_month', $lm) unless exists($monthsmap{$lm});
                my $fm_index = $monthsmap{$fm};
                my $lm_index = $monthsmap{$lm};

                my $is_range = (($lm eq $fm) && ($ld == $fd)) ? 0 : 1;
                my $year_offset = ($lm_index < $fm_index) || (($lm_index == $fm_index) && ($ld > $fd)) ? 1 : 0;

                #
                # Compute intersection between
                # [ start_when ; start_when+7 days ]
                # and [ $firstmonth ; $lastmonth ] (+/- 1 year)
                #
                foreach my $i (-1, 0, 1) {
                    my ($firstmonth, $lastmonth);
                    #
                    # Note on Feb 29 case for non-leap years:
                    # - if 'Feb 29' only, skip
                    # - if 'XXX YY - Feb 29', Feb 29 == Feb 28
                    # - if 'Feb 29 - XXX YY', Feb 29 == Mar 1
                    eval {
                        $firstmonth = timelocal(0, 0, 0, $fd, $fm_index, $startwhen_yo + $i);
                    };
                    if ($@) {
                        die $@ if ($fd != 29 || $fm_index != 1);
                        # Skip Feb 29 unless a range
                        next unless $is_range;
                        # Feb 29 => Mar 1 if a range
                        $firstmonth = timelocal(0, 0, 0, 1, 2, $startwhen_yo + $i);
                    }

                    if ($is_range) {
                        eval {
                            $lastmonth = timelocal(0, 0, 0, $ld, $lm_index, $startwhen_yo + $i + $year_offset);
                        };
                        if ($@) {
                            die $@ if ($ld != 29 || $lm_index != 1);
                            # Feb 29 => Feb 28
                            $lastmonth = timelocal(0, 0, 0, 28, 1, $startwhen_yo + $i + $year_offset);
                        }
                    } else {
                        $lastmonth = $firstmonth;
                    }
                    my ($inf, $sup) = (
                        max_ts($start_when, $firstmonth),
                        min_ts($start_when + $maxdays_ahead * 86400, $lastmonth)
                    );
                    for (my $t = $inf; $t <= $sup; $t += 86400) {
                        push(@ret, $t);
                    }
                }
            } elsif ($item =~ $datewithyear_regex) {
                # A date with year or range of dates with year
                my ($fm, $fd, $fy, $lm, $ld, $ly);
                if (defined($4)) {
                    ($fm, $fd, $fy, $lm, $ld, $ly) = ($1, $2, $3, $4, $5, $6);
                } else {
                    ($fm, $fd, $fy, $lm, $ld, $ly) = ($1, $2, $3, $1, $2, $3);
                }
                die i18n('program.invalid_month', $fm) unless exists($monthsmap{$fm});
                die i18n('program.invalid_month', $lm) unless exists($monthsmap{$lm});
                my $fm_index = $monthsmap{$fm};
                my $lm_index = $monthsmap{$lm};

                my $is_range = (($lm eq $fm) && ($ld == $fd) && ($ly == $fy)) ? 0 : 1;
                my $year_offset = ($lm_index < $fm_index) || (($lm_index == $fm_index) && ($ld < $fd)) ? 1 : 0;

                die i18n('program.invalid_yearrange', $item) if ($ly < $fy) || (($ly == $fy) && $year_offset);

                my ($firstdate, $lastdate);
                eval {
                    $firstdate = timelocal(0, 0, 0, $fd, $fm_index, $fy - 1900);
                };
                die i18n('program.invalid_date', $fm, $fd, $fy) if $@;

                eval {
                    $lastdate = timelocal(0, 0, 0, $ld, $lm_index, $ly - 1900);
                };
                die i18n('program.invalid_date', $lm, $ld, $ly) if $@;

                #
                # Compute intersection between
                # [ start_when ; start_when+7 days ]
                # and [ $firstdate ; $lastdate ]
                #
                my ($inf, $sup) = (
                    max_ts($start_when, $firstdate),
                    min_ts($start_when + $maxdays_ahead * 86400, $lastdate)
                );
                for (my $t = $inf; $t <= $sup; $t += 86400) {
                    push(@ret, $t);
                }
            } else {
                die i18n('program.unexpected_date_string', $item);
            }
        }
        die i18n('program.unexpected_date_string', $rest) unless $rest eq "";

        return uniq(sort { $a <=> $b } @ret);
    }

}

{
    our $_extract_times_regex = qr/^(\d\d):(\d\d)(?:\s*,\s*(.*))?$/;

    sub _extract_times {
        my ($timestr) = @_;

        my @ret = ();

        my $rest = $timestr;
        die i18n('program.unexpected_time_string', $rest) if $rest eq "";

        while ($rest =~ $_extract_times_regex) {
            my ($h, $m) = (int($1), int($2));
            die i18n('program.invalid_hour', $h) if $h >= 24;
            die i18n('program.invalid_minute', $m) if $m >= 60;
            $rest = defined($3) ? $3 : "";
            push(@ret, $h * 3600 + $m * 60);
        }
        die i18n('program.unexpected_time_string', $rest) unless $rest eq "";

        return uniq(sort { $a <=> $b } @ret);
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
