# Copyright (c) 2010-2015 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Scheduler;

use strict;
use warnings;

use base qw( AdvPowerMgr::Class );

use JSON qw( to_json );

use AdvPowerMgr::Config;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::Server;
use AdvPowerMgr::System;
use AdvPowerMgr::Utils qw( min_ts max_ts time2str );

__PACKAGE__->_mk_accessors(qw( config state ));
__PACKAGE__->_mk_accessors(qw( _reload ));

sub config_trigger {
    my $self = shift;
    my $was = shift;

    return unless defined($was);
    $self->_reload(1);
}

# state() is expected to return a hash with the following fields:
#
# 'phase' can be:
#  - 'idle'
#  - 'shutdown'
#  - 'suspended'
#
#  'wakeup_ts'
#  'shutdown_ts'
#
# 'nextupdate_ts' indicates next time some check will be performed
#
# In case where 'phase' == 'shutdown':
# '<plugin name>' fields indicate plugin state

sub new {
    my $class = shift;
    my $params = shift || {};

    die_param("config") unless defined($params->{config});

    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->config($params->{config});
    $self->state({});
    $self->_reload(0);
    return $self;
}

sub reload {
    my $self = shift;

    my $file = $self->config()->filename();
    my $config = AdvPowerMgr::Config->load({ filename => $file, use_defaults_on_failure => 1 });
    $self->config($config);
    return;
}

sub run {
    my $self = shift;

    my ($wakeup, $shutdown, $shutdown_phase, $nextupdate);
    my $wakeup_saved = undef;

RELOAD:
    my $config = $self->config();
    $self->_reload(0);

REINIT:
    my $state = {};
    $self->state($state);

    my $done = 0;
    while (!$done) {

        $nextupdate = 0;
        $shutdown = 0;
        $wakeup = 0;

        #
        # Compute Wakeup
        #

        foreach my $p ($config->prewakeup_plugins()) {
            my $res = $p->compute(); die_ie("Async!") unless defined($res);
            my $name = $p->name();
            $state->{$name} = $res;
            next unless $res->{enabled};
            info("Plugin '%s' reported wakeup not before '%s'", $name, time2str($res->{not_before_ts}));
            $wakeup = max_ts($wakeup, $res->{not_before_ts});
            $nextupdate = min_ts($nextupdate, $res->{nextupdate_ts});
        }

        if ($wakeup == -1) {
            info("Wakeup is suspended");
        } else {
            info("Wakeup not before '%s'", time2str($wakeup)) unless $wakeup == 0;
            my $res = $config->wakeup_plugin()->compute($wakeup);
            my $name = $config->wakeup_plugin()->name();
            $state->{$name} = $res;
            if ($res->{enabled}) {
                info("Plugin '%s' reported wakeup at '%s'", $name, time2str($res->{wakeup_ts}));
                $wakeup = $res->{wakeup_ts};
                $nextupdate = min_ts($nextupdate, $res->{nextupdate_ts});
            } else {
                info("No wakeup planned");
                $wakeup = 0;
            }
        }
        if (!defined($wakeup_saved) || ($wakeup != $wakeup_saved)) {
            AdvPowerMgr::System->set_wakeup($wakeup);
            $wakeup_saved = $wakeup;
        }

        #
        # Compute Shutdown
        #

        foreach my $p ($config->preshutdown_plugins(), $config->shutdown_plugin()) {
            my $res = $p->compute(); die_ie("Async!") unless defined($res);
            my $name = $p->name();
            $state->{$name} = $res;
            next unless $res->{enabled};
            info("Plugin '%s' reported shutdown at '%s'", $name, time2str($res->{shutdown_ts}));
            if ($res->{shutdown_ts} == -1) {
                $shutdown = -1;
            } else {
                $shutdown = min_ts($shutdown, $res->{shutdown_ts});
            }
            $nextupdate = min_ts($nextupdate, $res->{nextupdate_ts});
        }

        $shutdown_phase = $shutdown;
        $state->{phase} = 'idle';

        if ($shutdown == -1) {
            info("Shutdown is suspended");
            $state->{phase} = 'suspended';
        } elsif ($shutdown != 0) {
            my $offset = max_ts(map { $_->offset(); } $config->postshutdown_plugins());
            if ($offset != 0) {
                $shutdown_phase -= $offset;
                info("Adjusted shutdown phase to '%s' because of plugins", time2str($shutdown_phase));
            } else {
                info("Shutdown at '%s'", time2str($shutdown));
            }
            $nextupdate = min_ts($nextupdate, $shutdown_phase);
        } else {
            info("No shutdown planned");
        }

        if ($nextupdate == 0) {
            $nextupdate = -1;
        }

        # Sleep until next planned shutdown (or wakeup, for rearming purposes)
        if ($nextupdate == $shutdown_phase) {
            $done = 1;
        }
        $state->{nextupdate_ts} = int($nextupdate);
        $state->{wakeup_ts} = int($wakeup);
        $state->{shutdown_ts} = int($shutdown);

        info("Sleeping until '%s'", time2str($nextupdate));
        $self->_sleep_until($nextupdate);
        goto RELOAD if $self->_reload();
    }

    info("Reached shutdown phase at '%s'", time2str($shutdown_phase));
    $state->{phase} = 'shutdown';

    $done = 0;
    while (!$done) {
        # Planned shutdown reached, check whether it has been cancelled by a wakeup or not
        my $now = time();
        if ($wakeup != 0 && $now >= $wakeup && $now >= $shutdown) {
            # Shutdown has been cancelled by a wakeup
            info("Shutdown cancelled by an already reached wakeup '%s'", time2str($wakeup));
            goto REINIT;
        }

        $done = 1;
        $nextupdate = 0;

        my @plugins = $config->postshutdown_plugins();
        my @async_plugins = ();
        while (my $p = shift(@plugins)) {
            my $name = $p->name();
            info("Evaluating plugin '%s'", $name);
            my $res = $p->compute();
            unless (defined($res)) {
                # Re-enqueue the plugin in case of asynchronous computation
                die_ie("Async plugin '%s'", $name) if grep {/^$name$/} map { $_->name(); } @async_plugins;
                push(@async_plugins, $p);
                push(@plugins, $p);
                next; # while (plugins)
            }
            $state->{$name} = $res;
            next unless $res->{enabled};
            info("Shutdown plugin '%s' returned done=%d, nextupdate='%s'", $name,
                $res->{done}, time2str($res->{nextupdate_ts}));
            $done = 0 if !$res->{done};
            $nextupdate = min_ts($nextupdate, $res->{nextupdate_ts});
        }

        if ($nextupdate == 0) {
            $nextupdate = -1;
        }

        unless ($done) {
            $state->{nextupdate_ts} = int($nextupdate);
            info("Sleeping until '%s' before rechecking", time2str($nextupdate));
            $self->_sleep_until($nextupdate);
            goto RELOAD if $self->_reload();
        }
    }
    info("All done");

    my $server = AdvPowerMgr::Server->get();
    $server->stop() if defined($server);

    AdvPowerMgr::System->poweroff();
}

sub _sleep_until {
    my $self = shift;
    my $until = shift;

    die_ie("until == 0") if $until == 0;

    my $server = AdvPowerMgr::Server->get();
    unless (defined($server)) {
        while (1) {
            my $value;
            if ($until != -1) {
                $value = $until - time();
                return if $value <= 0;
            }
            select(undef, undef, undef, $value);
        }
    } else {
        $server->serve($until);
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
