# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Server;

use strict;
use warnings;

use base qw( AdvPowerMgr::Class );

use IO::Select;
use IO::Socket;
use Storable qw( fd_retrieve store_fd );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::System;
use AdvPowerMgr::Utils qw( millisleep );

__PACKAGE__->_mk_accessors(qw( socketpath timeout mode ));
__PACKAGE__->_mk_accessors(qw( _server _select _listener _services ));

sub socketpath_trigger {
    my $self = shift;
    my $was = shift;

    die_ie("UNSUPPORTED") if defined($was);
}

{
    our $singleton;

    sub server {
        my $class = shift;
        my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

        $params->{socketpath} = sprintf("%s/unixsocket", AdvPowerMgr::System->tmpdir()) unless defined($params->{socketpath});
        $params->{mode} = 0777 unless defined($params->{mode});
        $params->{timeout} = 1 unless defined($params->{timeout});
        my $path = $params->{socketpath};

        unless (defined($singleton)) {
            $singleton = $class->SUPER::new($params);
            bless($singleton, $class);
            $singleton->socketpath($path);
            $singleton->mode($params->{mode});
            $singleton->timeout($params->{timeout});
            $singleton->_server(1);
            $singleton->_setup_service();
        }
        die_ie("UNSUPPORTED") unless $singleton->socketpath() eq $path;
        return $singleton;
    }

    sub get {
        my $class = shift;
        return $singleton;
    }

    sub stop {
        my $self = shift;

        die "Internal error!" unless $self->_server();

        if (defined($self->_listener())) {
            close($self->_listener());
            unlink($self->socketpath());
        }
        undef($singleton);
    }
}

sub client {
    my $class = shift;
    my $params = defined($_[0]) ? { %{$_[0]} } : {}; shift;

    $params->{socketpath} = sprintf("%s/unixsocket", AdvPowerMgr::System->tmpdir()) unless defined($params->{socketpath});
    $params->{timeout} = 5 unless defined($params->{timeout});
    my $self = $class->SUPER::new($params);
    bless($self, $class);
    $self->socketpath($params->{socketpath});
    $self->timeout($params->{timeout});
    $self->_server(0);
    return $self;
}

sub call {
    my $self = ($_[0] eq __PACKAGE__) ? __PACKAGE__->client() : $_[0]; shift;
    my $method = shift or die_param("method");
    my @args = @_;

    die_ie("SERVER") if $self->_server();

    my $path = $self->socketpath();
    # No path -> assume server shut down gracefully
    return unless -e $path;

    my $sock;
    my $timeout = $self->timeout();
    eval {
        local $SIG{ALRM} = sub { CORE::die("timeout\n"); };
        alarm($timeout);
        while (1) {
            $sock = IO::Socket::UNIX->new(Type => SOCK_STREAM, Peer => $path);
            last if defined($sock);
            millisleep(1000);
        }
    };
    alarm(0);
    if ($@) {
        die_ie("failed to connect to server: " . ($@ eq "timeout\n" ? "Timeout" : "$!"));
    }

    my $data = { method => $method, args => [ @args ], wantarray => wantarray(), context => AdvPowerMgr::Log->context() }; 
    store_fd($data, $sock);
    $data = fd_retrieve($sock);
    close($sock);

    die $data->{error} if defined($data->{error});

    if (wantarray()) {
        return ( @{$data->{result}} );
    } else {
        return $data->{result};
    }
}

sub register {
    my $self = ($_[0] eq __PACKAGE__) ? __PACKAGE__->server() : $_[0]; shift;
    my $method = shift or die_param("method");
    my $handler = shift or die_param("handler");
    my $options = shift || {};

    foreach (keys(%$options)) {
        die_ie($_) unless $_ =~ m#^(context|bailout)$#;
    }
    die_ie("CLIENT") unless $self->_server();

    my $services = $self->_services();
    die_ie("'$method' already registered") if exists($services->{$method});
    $services->{$method} = { handler => $handler, %$options };
    info("Registered '%s' service", $method);

    return $self;
}

sub serve {
    my $self = shift;
    my $until = shift || -1; # 0: don't block, -1: forever

    die_ie("CLIENT") unless $self->_server();

    my $sel = $self->_select() or die_ie("no service");
    my $services = $self->_services();

    my $bailout = 0;

    while (!$bailout) {
        my $ready;
        if ($until == 0) {
            $ready = $sel->can_read(0);
            return unless $ready;
        } elsif ($until == -1) {
            $ready = $sel->can_read();
        } else {
            my $howlong = $until - time();
            return if $howlong <= 0;
            $ready = $sel->can_read($howlong);
        }
        next unless $ready;

        my $client;
        eval {
            $client = $self->_listener()->accept() or die_ie("'accept' failed: $!");
            my $data = fd_retrieve($client);
            my $wantarray = $data->{wantarray};
            my $method = $data->{method};
            my @args = @{$data->{args}};
            my $context = $data->{context};
            debug("Service '%s' is being called", $method);

            my $service = $services->{$method};

            die_ie("unexpected method '$method'") unless defined($service);

            my $oldcontext = AdvPowerMgr::Log->context($context);

            eval {
                my $ret;
                if ($wantarray) {
                    $ret = [ $service->{handler}->(@args) ];
                } else {
                    $ret = $service->{handler}->(@args);
                }
                store_fd({ result => $ret }, $client);
            };
            if ($@) {
                store_fd({ error => $@ }, $client);
            } elsif (defined($service->{bailout}) && $service->{bailout}) {
                debug("Requesting server to interrupt serving");
                $bailout = 1;
            }

            AdvPowerMgr::Log->context($oldcontext);
        };
        warning($@) if $@;
        close($client) if defined($client);
    }
}

sub _setup_service {
    my $self = shift;

    return $self if defined($self->_listener());
    my $sel = IO::Select->new();
    $self->_select($sel);

    my $path = $self->socketpath();
    unlink($path) if -e $path;
    my $listener = IO::Socket::UNIX->new(Type => SOCK_STREAM, Local => $path, Listen => 1, Reuse => 1);
    die_ie("Failed to create listener on file '$path': $!") unless defined($listener);
    $listener->timeout($self->timeout());
    chmod $self->mode(), $path;
    $self->_listener($listener);
    $sel->add($listener);
    $self->_services({});

    return $self;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
