#!/usr/bin/perl
# Copyright (c) 2010-2015 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::System::Linux;

use strict;
use warnings;

use base qw( AdvPowerMgr::System::Common );

use File::Path qw( mkpath );

use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::Utils qw( ipv4_to_hex time2str );

our $TMPDIR_FMT = "/var/tmp/%s";
our $STOREDIR_FMT = "%s/store";

our $ARPCACHE_FILE = "/proc/net/arp";
our $DISKSTATS_FILE = "/proc/diskstats";
our $NETSTAT_CMD = "/bin/netstat";
our $NETSTAT_FILE = "/proc/net/dev";

sub SMARTCTL_CMD {
    return "/usr/sbin/smartctl";
}

{
    our $get_arp_cache_regex = qr/^(\d+\.\d+\.\d+\.\d+)\s+\S+\s+\S+\s+([0-9a-f:]+)\s+/;

    sub get_arp_cache {
        my $class = shift;

        my %cache = ();

        open(my $fh, '<', $ARPCACHE_FILE) or die_open($ARPCACHE_FILE);
        while (my $line = <$fh>) {
            chomp($line);
            next unless $line =~ $get_arp_cache_regex;
            my ($mac, $ip) = ($2, $1);
            unless (exists($cache{$mac})) {
                $cache{$mac} = [];
            }
            push(@{$cache{$mac}}, $ip);
        }
        close($fh);

        return %cache;
    }
}

{
    our $get_networks_regex = qr/^(\d+\.\d+\.\d+\.\d+)\s+0\.0\.0\.0\s+(\d+\.\d+\.\d+\.\d+)\s+/;

    sub get_networks {
        my $class = shift;

        open(my $fh, '-|', $NETSTAT_CMD, "-rn") or die_exec($NETSTAT_CMD);
        my %net = (); # "Network in Hex" => "Mask in Hex" entries
        while (my $line = <$fh>) {
            next unless $line =~ $get_networks_regex;
            $net{ipv4_to_hex($1)} = ipv4_to_hex($2);
        }
        close($fh);

        return %net;
    }
}

{
    our $get_connections_regex_ipv4 = qr/^\S+\s+\S+\s+\S+\s+(\d+\.\d+\.\d+\.\d+):(\d+)\s+(\d+\.\d+\.\d+\.\d+):(\d+)\s+ESTABLISHED\s*$/;
    our $get_connections_regex_ipv4_in_ipv6 = qr/^\S+\s+\S+\s+\S+\s+::ffff:(\d+\.\d+\.\d+\.\d+)%\d+:(\d+)\s+::ffff:(\d+\.\d+\.\d+\.\d+):(\d+)\s+ESTABLISHED\s*$/;

    sub get_connections {
        my $class = shift;

        open(my $fh, '-|', $NETSTAT_CMD, "-tn") or die_exec($NETSTAT_CMD);
        my @res = ();
        while (my $line = <$fh>) {
            # Consider established IPv4 connections only
            if ($line =~ $get_connections_regex_ipv4) {
                push(@res, { local_ip => $1, local_port => int($2), remote_ip => $3, remote_port => int($4) });
            } elsif ($line =~ $get_connections_regex_ipv4_in_ipv6) {
                push(@res, { local_ip => $1, local_port => int($2), remote_ip => $3, remote_port => int($4) });
            }
        }
        close($fh);

        return @res;
    }
}

{
    our $get_network_stats_regex = qr/eth\d+:\s*(\d+)\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+(\d+)/;
    sub get_network_stats {
        my $class = shift;

        my $fh;
        open($fh, '<', $NETSTAT_FILE) or die_open($NETSTAT_FILE);
        my $traffic = 0;
        while (my $line = <$fh>) {
            next unless $line =~ $get_network_stats_regex;
            $traffic += $1 + $2;
        }
        close($fh);

        return $traffic;
    }
}

{
    our $get_disk_stats_md_regex = qr/^\s*\d+\s+\d+\s+md\d+\s+\d+\s+\d+\s+(\d+)\s+\d+\s+\d+\s+\d+\s+(\d+)/;
    our $get_disk_stats_sdX_regex = qr/^\s*\d+\s+\d+\s+sd[a-z]\s+\d+\s+\d+\s+(\d+)\s+\d+\s+\d+\s+\d+\s+(\d+)/;

    sub get_disk_stats {
        my $class = shift;

        open(my $fh, '<', $DISKSTATS_FILE) or die_open($DISKSTATS_FILE);
        my $traffic = 0;
        foreach my $regex ( $get_disk_stats_md_regex, $get_disk_stats_sdX_regex ) {
            while (my $line = <$fh>) {
                next unless $line =~ $regex;
                $traffic += ($1 + $2) * 512;
            }
            last if $traffic;
            seek($fh, 0, 0);
        }
        close($fh);

        return $traffic;
    }
}

{
    our @disks = ();
    our @disks_cmd = ();

    sub get_smartctl_disks {
        my $class = shift;

        unless (@disks) {
            open(my $fh, '-|', $class->SMARTCTL_CMD(), "--scan-open") or die_open($class->SMARTCTL_CMD());
            while (my $line = <$fh>) {
                $line =~ m/^(\/dev\/([a-z]+) .*) #/;
                push(@disks_cmd, $1);
                push(@disks, $2);
            }
            close($fh);
        }

        my @res = ();

        for (my $i = 0; $i < @disks; $i++) {
            # XXX Using autodetection doesn't work for some unknown reason
            # open(my $fh, "LANG=C " . $class->SMARTCTL_CMD() . " -c " . $disks_cmd[$i] . " |") or die_exec($class->SMARTCTL_CMD());
            # Force the use of 'ata' type for now
            open(my $fh, "LANG=C " . $class->SMARTCTL_CMD() . " -c /dev/" . $disks[$i] . " -d ata |") or die_exec($class->SMARTCTL_CMD());
            while (my $line = <$fh>) {
                next unless $line =~ m/Self-test routine in progress/;
                push(@res, $disks[$i]);
                last;
            }
            close($fh);
        }

        return @res;
    }
}

sub die_unless_root {
    debug_func_enter();
    return;
}

sub check_privilege {
    debug_func_enter();
    return ('SYNOTOKEN', 'ADMIN', 1);
}

sub hwid {
    return 'HWID';
}

{
    our $tmpdir;

    sub tmpdir {
        my $class = shift;

        if (!defined($tmpdir)) {
            $tmpdir = sprintf($TMPDIR_FMT, $class->_appname());
            -d $tmpdir || mkpath($tmpdir);
        }
        return $tmpdir;
    }
}

{
    our $storedir;

    sub storedir {
        my $class = shift;

        if (!defined($storedir)) {
            $storedir = sprintf($STOREDIR_FMT, $class->appdir());
            -d $storedir || mkpath($storedir);
        }
        return $storedir;
    }
}

sub lang {
    return undef;
}

sub set_wakeup {
    my $class = shift;
    my $when = shift;
    debug_func_enter({ when => time2str($when) });
    return;
}

sub poweroff {
    debug_func_enter();
    exit(0);
}

sub proxy {
    # No support for Proxy
    return undef;
}

sub notify {
    # No support for Notify
    return;
}

{
    our $testing;

    sub testing {
        my $class = shift;

        if (!defined($testing)) {
            $testing = -e sprintf("%s/testing", $class->storedir());
        }
        return $testing;
    }
}

sub init {
    return;
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
