# Copyright (c) 2010-2014 Renaud METRICH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package AdvPowerMgr::Validators;

use strict;
use warnings;

use base qw( AdvPowerMgr::Class );

use Time::Local qw( timelocal );

use JSON qw();

use AdvPowerMgr::I18N qw( i18n );

our @EXPORT_OK;
push(@EXPORT_OK, qw( boolean_validator positive_integer_validator strictly_positive_integer_validator date_validator ipaddr_validator macaddr_validator hostname_validator ));
our @EXPORT;
push(@EXPORT, @EXPORT_OK);

sub boolean_validator {
    my $class = shift if $_[0] eq __PACKAGE__;
    my $value = shift;

    if ($value =~ m#^(on|true)$# || $value eq JSON->true() || $value eq "1") {
        $value = 1;
    } elsif ($value =~ m#^(off|false)$# || $value eq JSON->false() || $value eq "0") {
        $value = 0;
    } else {
        die i18n('validators.not_boolean', $value);
    }

    return $value;
}

sub positive_integer_validator {
    my $class = shift if $_[0] eq __PACKAGE__;
    my $value = shift;

    die i18n('validators.not_positive_integer', $value) unless $value =~ m/^\d+$/;

    return $value;
}

sub strictly_positive_integer_validator {
    my $class = shift if $_[0] eq __PACKAGE__;
    my $value = shift;

    die i18n('validators.not_strictly_positive_integer', $value) unless $value =~ m/^\d+$/ && $value > 0;

    return $value;
}

sub date_validator {
    my $class = shift if $_[0] eq __PACKAGE__;
    my $value = shift;

    # YYYY-mm-dd
    die i18n('validators.not_en_date_format', $value) unless $value =~ m/^\d\d\d\d-\d\d-\d\d$/;
    my ($y, $m, $d) = split('-', $value);
    eval {
        timelocal(0, 0, 0, $d, $m - 1, $y - 1900);
    };
    die i18n('validators.invalid_date', $value) if $@;

    return $value;
}

{
    # Support for IPv4 only
    our $ipaddr_validator_regex = qr/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;

    sub ipaddr_validator {
        my $class = shift if $_[0] eq __PACKAGE__;
        my $value = shift;

        if ($value =~ $ipaddr_validator_regex) {
            goto DIE unless $1 > 0 && $1 < 255;
            goto DIE unless $2 >= 0 && $2 < 255;
            goto DIE unless $3 >= 0 && $3 < 255;
            goto DIE unless $4 > 0 && $4 < 255;
            return $value;
        }
    DIE:
        die i18n('validators.not_ip_addr', $value);
    }
}

{
    # Support for ethernet only
    our $macaddr_validator_regex = qr/^[0-9a-f][0-9a-f](?::[0-9a-f][0-9a-f]){5}$/i;

    sub macaddr_validator {
        my $class = shift if $_[0] eq __PACKAGE__;
        my $value = shift;

        die i18n('validators.not_mac_addr', $value) unless $value =~ $macaddr_validator_regex;
        return lc($value);
    }
}

{
    our $hostname_validator_regex = qr/^[a-z0-9](?:(?:[a-z0-9-])*[a-z0-9])?(?:\.[a-z](?:(?:[a-z0-9-])*[a-z0-9])?)*$/i;

    sub hostname_validator {
        my $class = shift if $_[0] eq __PACKAGE__;
        my $value = shift;

        die i18n('validators.not_hostname', $value) unless $value =~ $hostname_validator_regex;
        return lc($value);
    }
}

1;

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
