#!/usr/bin/perl

# Copyright (c) 2010-2015 Renaud METRICH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

use strict;
use warnings;

use FindBin qw( $Bin );
use lib "$Bin/../lib";

use AdvPowerMgr::Config;
use AdvPowerMgr::Daemon;
use AdvPowerMgr::Errors;
use AdvPowerMgr::Log;
use AdvPowerMgr::Log::File;
use AdvPowerMgr::Scheduler;
use AdvPowerMgr::Service::ConfigSave;
use AdvPowerMgr::Service::I18NReset;
use AdvPowerMgr::Service::I18NSave;
use AdvPowerMgr::Service::Log;
use AdvPowerMgr::Service::Monitor;
use AdvPowerMgr::Service::Notify;
use AdvPowerMgr::Service::SchedulerReload;
use AdvPowerMgr::Service::SuspendSave;
use AdvPowerMgr::System;
use AdvPowerMgr::Utils qw( time2str );

my $config;

eval {
    AdvPowerMgr::Log->context("Daemon");

    die "This program should not be started manually!" unless defined($ARGV[0]);

    AdvPowerMgr::System->die_unless_root();

    if ($ARGV[0] eq "start") {

        # start: DS is starting or package has been upgraded or manually stopped then restarted

        if (AdvPowerMgr::Daemon->exists()) {
            warning("An instance already exists, shouldn't be the case!");
            AdvPowerMgr::Daemon->kill();
        }

        AdvPowerMgr::Log->default(AdvPowerMgr::Log::File->new());

        info("NAS Hardware ID is '%s'", AdvPowerMgr::System->hwid());

        $config = AdvPowerMgr::Config->load({ use_defaults_on_failure => 1 });

        my $startup_file = sprintf("%s/startup", AdvPowerMgr::System->tmpdir());
        my $booting = (! -e $startup_file) ? 1 : 0;

        my $hib = $config->hibernate()->compute();
        $hib = $hib->{enabled} ? $hib->{not_before_ts} : 0;
        if (time() < $hib) {
            if ($booting && AdvPowerMgr::System->is_autobooting()) {
                info("Hibernation is active until '%s', powering off", time2str($hib));
                AdvPowerMgr::System->poweroff();
            } else {
                info("Hibernation is active but has been temporarily disabled since boot is manual or not currently booting");
            }
        }
        open(my $fh, '>', $startup_file) or die_create($startup_file); close($fh);

        my $sched = AdvPowerMgr::Scheduler->new({ config => $config });

        AdvPowerMgr::Daemon->daemonize(sub {
            AdvPowerMgr::Service::ConfigSave->register();
            AdvPowerMgr::Service::I18NReset->register();
            AdvPowerMgr::Service::I18NSave->register();
            AdvPowerMgr::Service::Log->register();
            AdvPowerMgr::Service::Monitor->register({ scheduler => $sched });
            AdvPowerMgr::Service::Notify->register();
            AdvPowerMgr::Service::SuspendSave->register();
            AdvPowerMgr::Service::SchedulerReload->register({ scheduler => $sched });

            $sched->run();
        });

    } elsif ($ARGV[0] eq "stop") {

        # DS is stopping, program next wakeup

        AdvPowerMgr::Daemon->kill();

        AdvPowerMgr::Log->default(AdvPowerMgr::Log::File->new());

        $config = AdvPowerMgr::Config->load({ use_defaults_on_failure => 1 });

        my $wakeup = $config->hibernate()->compute();
        $wakeup = $wakeup->{enabled} ? $wakeup->{not_before_ts} : 0;
        $wakeup = $config->poweron()->compute($wakeup);
        $wakeup = $wakeup->{enabled} ? $wakeup->{wakeup_ts} : 0;
        AdvPowerMgr::System->set_wakeup($wakeup);

    } elsif ($ARGV[0] eq "status") {

        my $status = AdvPowerMgr::Daemon->exists();
        exit(0) if $status == 1;
        exit(1) if $status == -1;
        exit(3) if $status == 0;
        die "Unexpected status '$status'!";

    } else {

        die "Invalid argument '$ARGV[0]'!";

    }
};

if ($@) {
    my $error = $@;
    chomp($error);
    fatal($error);
    AdvPowerMgr::System->notify($error);
    print STDERR $error, "\n";
    eval "require AdvPowerMgr::Home;";
    unless ($@) {
        AdvPowerMgr::Home->post_report($error, $config);
    }
    exit(1);
}

exit(0);

# vi:set autoindent tabstop=8 shiftwidth=4 softtabstop=4 expandtab:
